# -*- coding: utf-8 -*-

# Copyright 2015 Donne Martin. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import unicode_literals
from __future__ import print_function

import click

from .github import GitHub


click.disable_unicode_literals_warning = True
pass_github = click.make_pass_decorator(GitHub)


class GitHubCli(object):
    """The GitHubCli, builds `click` commands and runs `GitHub` methods."""

    @click.group()
    @click.pass_context
    def cli(ctx):
        """Main entry point for GitHubCli.

        :type ctx: :class:`click.core.Context`
        :param ctx: An instance of click.core.Context that stores an instance
            of `github.GitHub`.
        """
        # Create a GitHub object and remember it as the context object.
        # From this point onwards other commands can refer to it by using the
        # @pass_github decorator.
        ctx.obj = GitHub()

    @cli.command()
    @click.option('-e', '--enterprise', is_flag=True)
    @pass_github
    def configure(github, enterprise):
        """Configure gitsome.

        Attempts to authenticate the user and to set up the user's news feed.

        Usage/Example(s):
            gh configure
            gh configure -e
            gh configure --enterprise

        :type github: :class:`github.GitHub`
        :param github: An instance of `github.GitHub`.
        :type enterprise: bool
        :param enterprise: Determines whether to configure GitHub Enterprise.
            Default: False.
        """
        github.configure(enterprise)

    @cli.command('create-comment')
    @click.argument('user_repo_number')
    @click.option('-t', '--text')
    @pass_github
    def create_comment(github, user_repo_number, text):
        """Create a comment on the given issue.

        Usage:
            gh create-comment [user_repo_number] [-t/--text]

        Example(s):
            gh create-comment donnemartin/saws/1 -t "hello world"
            gh create-comment donnemartin/saws/1 --text "hello world"

        :type github: :class:`github.GitHub`
        :param github: An instance of `github.GitHub`.

        :type user_repo_number: str
        :param user_repo_number: The user/repo/issue_number.

        :type text: str
        :param text: The comment text.
        """
        github.create_comment(user_repo_number, text)

    @cli.command('create-issue')
    @click.argument('user_repo')
    @click.option('-t', '--issue_title')
    @click.option('-d', '--issue_desc', required=False)
    @pass_github
    def create_issue(github, user_repo, issue_title, issue_desc):
        """Create an issue.

        Usage:
            gh create-issue [user_repo] [-t/--issue_title] [-d/--issue_desc]

        Example(s):
            gh create-issue donnemartin/gitsome -t "title"
            gh create-issue donnemartin/gitsome -t "title" -d "desc"
            gh create-issue donnemartin/gitsome --issue_title "title" --issue_desc "desc"  # NOQA

        :type github: :class:`github.GitHub`
        :param github: An instance of `github.GitHub`.

        :type user_repo: str
        :param user_repo: The user/repo.

        :type issue_title: str
        :param issue_title: The issue title.

        :type issue_desc: str
        :param issue_desc: The issue body (optional).
        """
        github.create_issue(user_repo, issue_title, issue_desc)

    @cli.command('create-repo')
    @click.argument('repo_name')
    @click.option('-d', '--repo_desc', required=False)
    @click.option('-pr', '--private', is_flag=True)
    @pass_github
    def create_repo(github, repo_name, repo_desc, private):
        """Create a repo.

        Usage:
            gh create-repo [repo_name] [-d/--repo_desc] [-pr/--private]

        Example(s):
            gh create-repo repo_name
            gh create-repo repo_name -d "desc"
            gh create-repo repo_name --repo_desc "desc"
            gh create-repo repo_name -pr
            gh create-repo repo_name --repo_desc "desc" --private

        :type github: :class:`github.GitHub`
        :param github: An instance of `github.GitHub`.

        :type repo_name: str
        :param repo_name: The repo name.

        :type repo_desc: str
        :param repo_desc: The repo description (optional).

        :type private: bool
        :param private: Determines whether the repo is private.  Default: False.
        """
        github.create_repo(repo_name, repo_desc, private)

    @cli.command()
    @pass_github
    def emails(github):
        """List all the user's registered emails.

        Usage/Example(s):
            gh emails

        :type github: :class:`github.GitHub`
        :param github: An instance of `github.GitHub`.
        """
        github.emails()

    @cli.command()
    @click.option('-p', '--pager', is_flag=True)
    @pass_github
    def emojis(github, pager):
        """List all GitHub supported emojis.

        Usage:
            gh emojis [-p/--pager]

        Example(s):
            gh emojis | grep octo

        :type github: :class:`github.GitHub`
        :param github: An instance of `github.GitHub`.

        :type pager: bool
        :param pager: Determines whether to show the output in a pager,
            if available.
        """
        github.emojis(pager)

    @cli.command()
    @click.argument('user_or_repo', required=False, default='')
    @click.option('-pr', '--private', is_flag=True, default=False)
    @click.option('-p', '--pager', is_flag=True)
    @pass_github
    def feed(github, user_or_repo, private, pager):
        """List all activity for the given user or repo.

        If `user_or_repo` is not provided, uses the logged in user's news feed
        seen while visiting https://github.com.  If `user_or_repo` is provided,
        shows either the public or `[-pr/--private]` feed activity of the user
        or repo.

        Usage:
            gh feed [user_or_repo] [-pr/--private] [-p/--pager]

        Examples:
            gh feed
            gh feed | grep foo
            gh feed donnemartin
            gh feed donnemartin -pr -p
            gh feed donnemartin --private --pager
            gh feed donnemartin/haxor-news -p

        :type github: :class:`github.GitHub`
        :param github: An instance of `github.GitHub`.

        :type user_or_repo: str
        :param user_or_repo: The user or repo to list events for (optional).
            If no entry, defaults to the logged in user's feed.

        :type private: bool
        :param private: Determines whether to show the private events (True)
            or public events (False).

        :type pager: bool
        :param pager: Determines whether to show the output in a pager,
            if available.
        """
        github.feed(user_or_repo, private, pager)

    @cli.command()
    @click.argument('user', required=False)
    @click.option('-p', '--pager', is_flag=True)
    @pass_github
    def followers(github, user, pager):
        """List all followers and the total follower count.

        Usage:
            gh followers [user] [-p/--pager]

        Example(s):
            gh followers
            gh followers -p
            gh followers octocat --pager

        :type github: :class:`github.GitHub`
        :param github: An instance of `github.GitHub`.

        :type user: str
        :param user: The user login (optional).
            If None, returns the followers of the logged in user.

        :type pager: bool
        :param pager: Determines whether to show the output in a pager,
            if available.
        """
        github.followers(user, pager)

    @cli.command()
    @click.argument('user', required=False)
    @click.option('-p', '--pager', is_flag=True)
    @pass_github
    def following(github, user, pager):
        """List all followed users and the total followed count.

        Usage:
            gh following [user] [-p/--pager]

        Example(s):
            gh following
            gh following -p
            gh following octocat --pager

        :type github: :class:`github.GitHub`
        :param github: An instance of `github.GitHub`.

        :type user: str
        :param user: The user login.
            If None, returns the followed users of the logged in user.

        :type pager: bool
        :param pager: Determines whether to show the output in a pager,
            if available.
        """
        github.following(user, pager)

    @cli.command('gitignore-template')
    @click.argument('language')
    @pass_github
    def gitignore_template(github, language):
        """Output the gitignore template for the given language.

        Usage:
            gh gitignore-template [language]

        Example(s):
            gh gitignore-template Python
            gh gitignore-template Python > .gitignore

        :type github: :class:`github.GitHub`
        :param github: An instance of `github.GitHub`.

        :type language: str
        :param language: The language.
        """
        github.gitignore_template(language)

    @cli.command('gitignore-templates')
    @click.option('-p', '--pager', is_flag=True)
    @pass_github
    def gitignore_templates(github, pager):
        """Output all supported gitignore templates.

        Usage:
            gh gitignore-templates

        Example(s):
            gh gitignore-templates
            gh gitignore-templates -p
            gh gitignore-templates --pager

        :type github: :class:`github.GitHub`
        :param github: An instance of `github.GitHub`.

        :type pager: bool
        :param pager: Determines whether to show the output in a pager,
            if available.
        """
        github.gitignore_templates(pager)

    @cli.command()
    @click.argument('user_repo_number')
    @pass_github
    def issue(github, user_repo_number):
        """Output detailed information about the given issue.

        Usage:
            gh issue [user_repo_number]

        Example(s):
            gh issue donnemartin/saws/1

        :type github: :class:`github.GitHub`
        :param github: An instance of `github.GitHub`.

        :type user_repo_number: str
        :param user_repo_number: The user/repo/issue_number.
        """
        github.issue(user_repo_number)

    @cli.command()
    @click.option('-f', '--issue_filter', required=False, default='subscribed')
    @click.option('-s', '--issue_state', required=False, default='open')
    @click.option('-l', '--limit', required=False, default=1000)
    @click.option('-p', '--pager', is_flag=True)
    @pass_github
    def issues(github, issue_filter, issue_state, limit, pager):
        """List all issues matching the filter.

        Usage:
            gh issues [-f/--issue_filter] [-s/--issue_state] [-l/--limit] [-p/--pager]  # NOQA

        Example(s):
            gh issues
            gh issues -f assigned
            gh issues ---issue_filter created
            gh issues -s all -l 20 -p
            gh issues --issue_state closed --limit 20 --pager
            gh issues -f created -s all -p

        :type github: :class:`github.GitHub`
        :param github: An instance of `github.GitHub`.

        :type issue_filter: str
        :param issue_filter: assigned, created, mentioned, subscribed (default).

        :type issue_state: str
        :param issue_state: all, open (default), closed.

        :type limit: int
        :param limit: The number of items to display.

        :type pager: bool
        :param pager: Determines whether to show the output in a pager,
            if available.
        """
        github.issues_setup(issue_filter, issue_state, limit, pager)

    @cli.command()
    @click.argument('license_name')
    @pass_github
    def license(github, license_name):
        """Output the license template for the given license.

        Usage:
            gh license [license_name]

        Example(s):
            gh license apache-2.0
            gh license mit > LICENSE

        :type github: :class:`github.GitHub`
        :param github: An instance of `github.GitHub`.

        :type license_name: str
        :param license_name: The license name.
        """
        github.license(license_name)

    @cli.command()
    @pass_github
    def licenses(github):
        """Output all supported license templates.

        Usage/Example(s):
            gh licenses

        :type github: :class:`github.GitHub`
        :param github: An instance of `github.GitHub`.
        """
        github.licenses()

    @cli.command()
    @click.option('-b', '--browser', is_flag=True)
    @click.option('-t', '--text_avatar', is_flag=True)
    @click.option('-l', '--limit', required=False, default=1000)
    @click.option('-p', '--pager', is_flag=True)
    @pass_github
    def me(github, browser, text_avatar, limit, pager):
        """List information about the logged in user.

        Usage:
            gh me [-b/--browser] [-t/--text_avatar] [-l/--limit] [-p/--pager]

        Example(s):
            gh me
            gh me -b
            gh me --browser
            gh me -t -l 20 -p
            gh me --text_avatar --limit 20 --pager

        :type github: :class:`github.GitHub`
        :param github: An instance of `github.GitHub`.

        :type browser: bool
        :param browser: Determines whether to view the profile
            in a browser, or in the terminal.

        :type text_avatar: bool
        :param text_avatar: Determines whether to view the profile
            avatar in plain text instead of ansi (default).
            On Windows this value is always set to True due to lack of
            support of `img2txt` on Windows.

        :type limit: int
        :param limit: The number of user repos to display.

        :type pager: bool
        :param pager: Determines whether to show the output in a pager,
            if available.
        """
        github.user_me(browser, text_avatar, limit, pager)

    @cli.command()
    @click.option('-l', '--limit', required=False, default=1000)
    @click.option('-p', '--pager', is_flag=True)
    @pass_github
    def notifications(github, limit, pager):
        """List all notifications.

        Usage:
            gh notifications [-l/--limit] [-p/--pager]

        Example(s):
            gh notifications
            gh notifications -l 20 -p
            gh notifications --limit 20 --pager

        :type github: :class:`github.GitHub`
        :param github: An instance of `github.GitHub`.

        :type limit: int
        :param limit: The number of items to display.

        :type pager: bool
        :param pager: Determines whether to show the output in a pager,
            if available.
        """
        github.notifications(limit, pager)

    @cli.command('octo')
    @click.argument('say', required=False)
    @pass_github
    def octocat(github, say):
        """Output an Easter egg or the given message from Octocat.

        Usage:
            gh octo [say]

        Example(s):
            gh octo
            gh octo "foo bar"

        :type github: :class:`github.GitHub`
        :param github: An instance of `github.GitHub`.

        :type say: str
        :param say: What Octocat should say.
                If say is None, octocat speaks an Easter egg.
        """
        github.octocat(say)

    @cli.command('pull-request')
    @click.argument('user_repo_number')
    @pass_github
    def pull_request(github, user_repo_number):
        """Output detailed information about the given pull request.

        Usage:
            gh pull-request [user_repo_number]

        Example(s):
            gh pull-request donnemartin/saws/80

        :type github: :class:`github.GitHub`
        :param github: An instance of `github.GitHub`.

        :type user_repo_number: str
        :param user_repo_number: The user/repo/pull_number.
        """
        github.issue(user_repo_number)

    @cli.command('pull-requests')
    @click.option('-l', '--limit', required=False, default=1000)
    @click.option('-p', '--pager', is_flag=True)
    @pass_github
    def pull_requests(github, limit, pager):
        """List all pull requests.

        Usage:
            gh pull-requests [-l/--limit] [-p/--pager]

        Example(s):
            gh pull-requests
            gh pull-requests -l 20 -p
            gh pull-requests --limit 20 --pager

        :type github: :class:`github.GitHub`
        :param github: An instance of `github.GitHub`.

        :type limit: int
        :param limit: The number of items to display.

        :type pager: bool
        :param pager: Determines whether to show the output in a pager,
            if available.
        """
        github.pull_requests(limit, pager)

    @cli.command('rate-limit')
    @pass_github
    def rate_limit(github):
        """Output the rate limit.  Not available for GitHub Enterprise.

        Logged in users can make 5000 requests per hour.
        See: https://developer.github.com/v3/#rate-limiting

        Usage/Example(s):
            gh rate-limit

        :type github: :class:`github.GitHub`
        :param github: An instance of `github.GitHub`.
        """
        github.rate_limit()

    @cli.command('repos')
    @click.argument('repo_filter', required=False, default='')
    @click.option('-l', '--limit', required=False, default=1000)
    @click.option('-p', '--pager', is_flag=True)
    @pass_github
    def repositories(github, repo_filter, limit, pager):
        """List all repos matching the given filter.

        Usage:
            gh repos [repo_filter] [-l/--limit] [-p/--pager]

        Example(s):
            gh repos
            gh repos "data-science"
            gh repos "data-science" -l 20 -p
            gh repos "data-science" --limit 20 --pager

        :type github: :class:`github.GitHub`
        :param github: An instance of `github.GitHub`.

        :type repo_filter: str
        :param repo_filter:  The filter for repo names.
            Only repos matching the filter will be returned.
            If None, outputs all the logged in user's repos.

        :type limit: int
        :param limit: The number of items to display.

        :type pager: bool
        :param pager: Determines whether to show the output in a pager,
            if available.
        """
        github.repositories_setup(repo_filter, limit, pager)

    @cli.command('repo')
    @click.argument('user_repo')
    @pass_github
    def repository(github, user_repo):
        """Output detailed information about the given repo.

        Usage:
            gh repo [user_repo]

        Example(s):
            gh repo donnemartin/gitsome

        :type github: :class:`github.GitHub`
        :param github: An instance of `github.GitHub`.

        :type user_repo: str
        :param user_repo: The user/repo.
        """
        github.repository(user_repo)

    @cli.command('search-issues')
    @click.argument('query')
    @click.option('-l', '--limit', required=False, default=1000)
    @click.option('-p', '--pager', is_flag=True)
    @pass_github
    def search_issues(github, query, limit, pager):
        """Search for all issues matching the given query.

        Usage:
            gh search-issues [query] [-l/--limit] [-p/--pager]

        Example(s):
            gh search-issues "foo type:pr author:donnemartin" -l 20 -p
            gh search-issues "foobarbaz in:title created:>=2015-01-01" --limit 20 --pager  # NOQA

        Additional Example(s):
            Search issues that have your user name tagged @donnemartin:
                gh search-issues "is:issue donnemartin is:open" -p

            Search issues that have the most +1s:
                gh search-issues "is:open is:issue sort:reactions-+1-desc" -p

            Search issues that have the most comments:
                gh search-issues "is:open is:issue sort:comments-desc" -p

            Search issues with the "help wanted" tag:
                gh search-issues "is:open is:issue label:\"help wanted\"" -p

            Search all your open private issues:
                gh search-issues "is:open is:issue is:private" -p

        The query can contain any combination of the following supported
        qualifers:

        - `type` With this qualifier you can restrict the search to issues
          or pull request only.
        - `in` Qualifies which fields are searched. With this qualifier you
          can restrict the search to just the title, body, comments, or any
          combination of these.
        - `author` Finds issues created by a certain user.
        - `assignee` Finds issues that are assigned to a certain user.
        - `mentions` Finds issues that mention a certain user.
        - `commenter` Finds issues that a certain user commented on.
        - `involves` Finds issues that were either created by a certain user,
          assigned to that user, mention that user, or were commented on by
          that user.
        - `state` Filter issues based on whether they’re open or closed.
        - `labels` Filters issues based on their labels.
        - `language` Searches for issues within repositories that match a
          certain language.
        - `created` or `updated` Filters issues based on times of creation,
          or when they were last updated.
        - `comments` Filters issues based on the quantity of comments.
        - `user` or `repo` Limits searches to a specific user or
          repository.

        For more information about these qualifiers, see: http://git.io/d1oELA

        :type github: :class:`github.GitHub`
        :param github: An instance of `github.GitHub`.

        :type query: str
        :param query: The search query.

        :type limit: int
        :param limit: The number of items to display.

        :type pager: bool
        :param pager: Determines whether to show the output in a pager,
            if available.
        """
        github.search_issues(query, limit, pager)

    @cli.command('search-repos')
    @click.argument('query')
    @click.option('-s', '--sort', required=False, default='')
    @click.option('-l', '--limit', required=False, default=1000)
    @click.option('-p', '--pager', is_flag=True)
    @pass_github
    def search_repositories(github, query, sort, limit, pager):
        """Search for all repos matching the given query.

        Usage:
            gh search-repos [query] [-s/--sort] [-l/--limit] [-p/--pager]

        Example(s):
            gh search-repos "maps language:python" -s stars -l 20 -p
            gh search-repos "created:>=2015-01-01 stars:>=1000 language:python" --sort stars --limit 20 --pager  # NOQA

        The query can contain any combination of the following supported
        qualifers:

        - `in` Qualifies which fields are searched. With this qualifier you
          can restrict the search to just the repository name, description,
          readme, or any combination of these.
        - `size` Finds repositories that match a certain size (in
          kilobytes).
        - `forks` Filters repositories based on the number of forks, and/or
          whether forked repositories should be included in the results at
          all.
        - `created` or `pushed` Filters repositories based on times of
          creation, or when they were last updated. Format: `YYYY-MM-DD`.
          Examples: `created:<2011`, `pushed:<2013-02`,
          `pushed:>=2013-03-06`
        - `user` or `repo` Limits searches to a specific user or
          repository.
        - `language` Searches repositories based on the language they're
          written in.
        - `stars` Searches repositories based on the number of stars.

        For more information about these qualifiers, see: http://git.io/4Z8AkA

        :type github: :class:`github.GitHub`
        :param github: An instance of `github.GitHub`.

        :type query: str
        :param query: The search query.

        :type sort: str
        :param sort: Optional: 'stars', 'forks', 'updated'.
            If not specified, sorting is done by query best match.

        :type limit: int
        :param limit: The number of items to display.

        :type pager: bool
        :param pager: Determines whether to show the output in a pager,
            if available.
        """
        github.search_repositories(query, sort, limit, pager)

    @cli.command()
    @click.argument('repo_filter', required=False, default='')
    @click.option('-l', '--limit', required=False, default=1000)
    @click.option('-p', '--pager', is_flag=True)
    @pass_github
    def starred(github, repo_filter, limit, pager):
        """Output starred repos.

        Usage:
            gh starred [repo_filter] [-l/--limit] [-p/--pager]

        Example(s):
            gh starred
            gh starred foo -l 20 -p
            gh starred foo --limit 20 --pager

        :type github: :class:`github.GitHub`
        :param github: An instance of `github.GitHub`.

        :type repo_filter: str
        :param repo_filter:  The filter for repo names.
            Only repos matching the filter will be returned.
            If None, outputs all starred repos.

        :type limit: int
        :param limit: The number of items to display.

        :type pager: bool
        :param pager: Determines whether to show the output in a pager,
            if available.
        """
        github.starred(repo_filter, limit, pager)

    @cli.command()
    @click.argument('language', required=False, default='Overall')
    @click.option('-w', '--weekly', is_flag=True)
    @click.option('-m', '--monthly', is_flag=True)
    @click.option('-D', '--devs', is_flag=True)
    @click.option('-b', '--browser', is_flag=True)
    @click.option('-p', '--pager', is_flag=True)
    @pass_github
    def trending(github, language, weekly, monthly, devs, browser, pager):
        """List trending repos for the given language.

        Usage:
            gh trending [language] [-w/--weekly] [-m/--monthly] [-D/--devs] [-b/--browser] [-p/--pager]  # NOQA

        Example(s):
            gh trending
            gh trending Python -w -p
            gh trending Python --weekly --devs --browser
            gh trending --browser

        :type github: :class:`github.GitHub`
        :param github: An instance of `github.GitHub`.

        :type language: str
        :param language: The language (optional).
            If blank, shows 'Overall'.

        :type weekly: bool
        :param weekly: Determines whether to show the weekly rankings.
            Daily is the default.

        :type monthly: bool
        :param monthly: Determines whether to show the monthly rankings.
            Daily is the default.
            If both `monthly` and `weekly` are set, `monthly` takes precedence.

        :type devs: bool
        :param devs: determines whether to display the trending
                devs or repos.  Only valid with the -b/--browser option.

        :type browser: bool
        :param browser: Determines whether to view the profile
                in a browser, or in the terminal.

        :type pager: bool
        :param pager: Determines whether to show the output in a pager,
            if available.
        """
        github.trending(language, weekly, monthly, devs, browser, pager)

    @cli.command()
    @click.argument('user_id', required=True)
    @click.option('-b', '--browser', is_flag=True)
    @click.option('-t', '--text_avatar', is_flag=True)
    @click.option('-l', '--limit', required=False, default=1000)
    @click.option('-p', '--pager', is_flag=True)
    @pass_github
    def user(github, user_id, browser, text_avatar, limit, pager):
        """List information about the given user.

        Usage:
            gh user [user_id] [-b/--browser] [-t/--text_avatar] [-l/--limit] [-p/--pager]  # NOQA

        Example(s):
            gh user octocat
            gh user octocat -b
            gh user octocat --browser
            gh user octocat -t -l 10 -p
            gh user octocat --text_avatar --limit 10 --pager

        :type github: :class:`github.GitHub`
        :param github: An instance of `github.GitHub`.

        :type user_id: str
        :param user_id: The user id/login.
            If None, returns followers of the logged in user.

        :type browser: bool
        :param browser: Determines whether to view the profile
            in a browser, or in the terminal.

        :type text_avatar: bool
        :param text_avatar: Determines whether to view the profile
            avatar in plain text instead of ansi (default).
            On Windows this value is always set to True due to lack of
            support of `img2txt` on Windows.

        :type limit: int
        :param limit: The number of items to display.

        :type pager: bool
        :param pager: Determines whether to show the output in a pager,
            if available.
        """
        github.user(user_id, browser, text_avatar, limit, pager)

    @cli.command()
    @click.argument('index')
    @click.option('-b', '--browser', is_flag=True)
    @pass_github
    def view(github, index, browser):
        """View the given notification/repo/issue/pull_request/user index.

        This method is meant to be called after one of the following commands
        which outputs a table of notifications/repos/issues/pull_requests/users:

            gh repos
            gh search_repos
            gh starred

            gh issues
            gh pull_requests
            gh search_issues

            gh notifications
            gh trending

            gh user
            gh me

        Usage:
            gh view [index] [-b/--browser]

        Example(s):
            gh repos
            gh view 1

            gh starred
            gh view 1 -b
            gh view 1 --browser

        :type github: :class:`github.GitHub`
        :param github: An instance of `github.GitHub`.

        :type index: str
        :param index: Determines the index to view.

        :type browser: bool
        :param browser: Determines whether to view the profile
            in a browser, or in the terminal.
        """
        github.view(int(index), browser)



# -*- coding: utf-8 -*-
"""Prompt formatter for simple version control branches"""
# pylint:disable=no-member, invalid-name

import os
import sys
import queue
import builtins
import threading
import subprocess

import xonsh.tools as xt


def _get_git_branch(q):
    denv = builtins.__xonsh__.env.detype()
    try:
        branches = xt.decode_bytes(
            subprocess.check_output(
                ["git", "branch"], env=denv, stderr=subprocess.DEVNULL
            )
        ).splitlines()
    except (subprocess.CalledProcessError, OSError, FileNotFoundError):
        q.put(None)
    else:
        for branch in branches:
            if not branch.startswith("* "):
                continue
            elif branch.endswith(")"):
                branch = branch.split()[-1][:-1]
            else:
                branch = branch.split()[-1]

            q.put(branch)
            break
        else:
            q.put(None)


def get_git_branch():
    """Attempts to find the current git branch. If this could not
    be determined (timeout, not in a git repo, etc.) then this returns None.
    """
    branch = None
    timeout = builtins.__xonsh__.env.get("VC_BRANCH_TIMEOUT")
    q = queue.Queue()

    t = threading.Thread(target=_get_git_branch, args=(q,))
    t.start()
    t.join(timeout=timeout)
    try:
        branch = q.get_nowait()
    except queue.Empty:
        branch = None
    return branch


def _get_hg_root(q):
    _curpwd = builtins.__xonsh__.env["PWD"]
    while True:
        if not os.path.isdir(_curpwd):
            return False
        if any([b.name == ".hg" for b in xt.scandir(_curpwd)]):
            q.put(_curpwd)
            break
        else:
            _oldpwd = _curpwd
            _curpwd = os.path.split(_curpwd)[0]
            if _oldpwd == _curpwd:
                return False


def get_hg_branch(root=None):
    """Try to get the mercurial branch of the current directory,
    return None if not in a repo or subprocess.TimeoutExpired if timed out.
    """
    env = builtins.__xonsh__.env
    timeout = env["VC_BRANCH_TIMEOUT"]
    q = queue.Queue()
    t = threading.Thread(target=_get_hg_root, args=(q,))
    t.start()
    t.join(timeout=timeout)
    try:
        root = q.get_nowait()
    except queue.Empty:
        return None
    if env.get("VC_HG_SHOW_BRANCH"):
        # get branch name
        branch_path = os.path.sep.join([root, ".hg", "branch"])
        if os.path.exists(branch_path):
            with open(branch_path, "r") as branch_file:
                branch = branch_file.read()
        else:
            branch = "default"
    else:
        branch = ""
    # add bookmark, if we can
    bookmark_path = os.path.sep.join([root, ".hg", "bookmarks.current"])
    if os.path.exists(bookmark_path):
        with open(bookmark_path, "r") as bookmark_file:
            active_bookmark = bookmark_file.read()
        if env.get("VC_HG_SHOW_BRANCH") is True:
            branch = "{0}, {1}".format(
                *(b.strip(os.linesep) for b in (branch, active_bookmark))
            )
        else:
            branch = active_bookmark.strip(os.linesep)
    else:
        branch = branch.strip(os.linesep)
    return branch


_FIRST_BRANCH_TIMEOUT = True


def _first_branch_timeout_message():
    global _FIRST_BRANCH_TIMEOUT
    sbtm = builtins.__xonsh__.env["SUPPRESS_BRANCH_TIMEOUT_MESSAGE"]
    if not _FIRST_BRANCH_TIMEOUT or sbtm:
        return
    _FIRST_BRANCH_TIMEOUT = False
    print(
        "xonsh: branch timeout: computing the branch name, color, or both "
        "timed out while formatting the prompt. You may avoid this by "
        "increasing the value of $VC_BRANCH_TIMEOUT or by removing branch "
        "fields, like {curr_branch}, from your $PROMPT. See the FAQ "
        "for more details. This message will be suppressed for the remainder "
        "of this session. To suppress this message permanently, set "
        "$SUPPRESS_BRANCH_TIMEOUT_MESSAGE = True in your xonshrc file.",
        file=sys.stderr,
    )


def current_branch():
    """Gets the branch for a current working directory. Returns an empty string
    if the cwd is not a repository.  This currently only works for git and hg
    and should be extended in the future.  If a timeout occurred, the string
    '<branch-timeout>' is returned.
    """
    branch = None
    cmds = builtins.__xonsh__.commands_cache
    # check for binary only once
    if cmds.is_empty():
        has_git = bool(cmds.locate_binary("git", ignore_alias=True))
        has_hg = bool(cmds.locate_binary("hg", ignore_alias=True))
    else:
        has_git = bool(cmds.lazy_locate_binary("git", ignore_alias=True))
        has_hg = bool(cmds.lazy_locate_binary("hg", ignore_alias=True))
    if has_git:
        branch = get_git_branch()
    if not branch and has_hg:
        branch = get_hg_branch()
    if isinstance(branch, subprocess.TimeoutExpired):
        branch = "<branch-timeout>"
        _first_branch_timeout_message()
    return branch or None


def _git_dirty_working_directory(q, include_untracked):
    status = None
    denv = builtins.__xonsh__.env.detype()
    try:
        cmd = ["git", "status", "--porcelain"]
        if include_untracked:
            cmd.append("--untracked-files=normal")
        else:
            cmd.append("--untracked-files=no")
        status = subprocess.check_output(cmd, stderr=subprocess.DEVNULL, env=denv)
    except (subprocess.CalledProcessError, OSError, FileNotFoundError):
        q.put(None)
    if status is not None:
        return q.put(bool(status))


def git_dirty_working_directory(include_untracked=False):
    """Returns whether or not the git directory is dirty. If this could not
    be determined (timeout, file not found, etc.) then this returns None.
    """
    timeout = builtins.__xonsh__.env.get("VC_BRANCH_TIMEOUT")
    q = queue.Queue()
    t = threading.Thread(
        target=_git_dirty_working_directory, args=(q, include_untracked)
    )
    t.start()
    t.join(timeout=timeout)
    try:
        return q.get_nowait()
    except queue.Empty:
        return None


def hg_dirty_working_directory():
    """Computes whether or not the mercurial working directory is dirty or not.
    If this cannot be determined, None is returned.
    """
    env = builtins.__xonsh__.env
    cwd = env["PWD"]
    denv = env.detype()
    vcbt = env["VC_BRANCH_TIMEOUT"]
    # Override user configurations settings and aliases
    denv["HGRCPATH"] = ""
    try:
        s = subprocess.check_output(
            ["hg", "identify", "--id"],
            stderr=subprocess.PIPE,
            cwd=cwd,
            timeout=vcbt,
            universal_newlines=True,
            env=denv,
        )
        return s.strip(os.linesep).endswith("+")
    except (
        subprocess.CalledProcessError,
        subprocess.TimeoutExpired,
        FileNotFoundError,
    ):
        return None


def dirty_working_directory():
    """Returns a boolean as to whether there are uncommitted files in version
    control repository we are inside. If this cannot be determined, returns
    None. Currently supports git and hg.
    """
    dwd = None
    cmds = builtins.__xonsh__.commands_cache
    if cmds.lazy_locate_binary("git", ignore_alias=True):
        dwd = git_dirty_working_directory()
    if cmds.lazy_locate_binary("hg", ignore_alias=True) and dwd is None:
        dwd = hg_dirty_working_directory()
    return dwd


def branch_color():
    """Return red if the current branch is dirty, yellow if the dirtiness can
    not be determined, and green if it clean. These are bold, intense colors
    for the foreground.
    """
    dwd = dirty_working_directory()
    if dwd is None:
        color = "{YELLOW}"
    elif dwd:
        color = "{RED}"
    else:
        color = "{GREEN}"
    return color


def branch_bg_color():
    """Return red if the current branch is dirty, yellow if the dirtiness can
    not be determined, and green if it clean. These are background colors.
    """
    dwd = dirty_working_directory()
    if dwd is None:
        color = "{BACKGROUND_YELLOW}"
    elif dwd:
        color = "{BACKGROUND_RED}"
    else:
        color = "{BACKGROUND_GREEN}"
    return color



# -*- coding: utf-8 -*-

# Copyright 2015 Donne Martin. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import unicode_literals
from __future__ import print_function

import os
import platform
import sys
import urllib
import webbrowser

from .lib.github3 import null
from .lib.github3.exceptions import AuthenticationFailed, UnprocessableEntity
from .lib.img2txt import img2txt
import click
import feedparser
from requests.exceptions import MissingSchema, SSLError

from .config import Config
from .formatter import Formatter
from .rss_feed import language_rss_map
from .table import Table
from .view_entry import ViewEntry
from .web_viewer import WebViewer
from .utils import TextUtils


class GitHub(object):
    """Provide integration with the GitHub API.

    :type config: :class:`config.Config`
    :param config: An instance of `config.Config`.

    :type formatter: :class:`formatter.Formatter`
    :param formatter: An instance of `formatter.Formatter`.

    :type img2txt: callable
    :param img2txt: A callable fom img2txt.

    :type table: :class:`table.Table`
    :param table: An instance of `table.Table`.

    :type trend_parser: :class:`feedparser`
    :param trend_parser: An instance of `feedparser`.

    :type web_viewer: :class:`web_viewer.WebViewer`
    :param web_viewer: An instance of `web_viewer.WebViewer`.

    :type _base_url: str
    :param _base_url: The base GitHub or GitHub Enterprise url.
    """

    def __init__(self):
        self.config = Config()
        self.formatter = Formatter(self.config)
        self.img2txt = img2txt.img2txt
        self.table = Table(self.config)
        self.web_viewer = WebViewer(self.config)
        self.trend_parser = feedparser
        self.text_utils = TextUtils()
        self._base_url = 'https://github.com/'

    @property
    def base_url(self):
        return self.config.enterprise_url or self._base_url

    def add_base_url(self, url):
        """Add the base url if it is not already part of the given url.

        :type url: str
        :param url: The url.

        :return: The url including the base url.
        """
        return self.base_url + url if self.base_url not in url else url

    def authenticate(func):
        """Decorator that authenticates credentials.

        :type func: callable
        :param func: A method to execute if authorization passes.

        :return: The return value of `func` if authorization passes, or
            None if authorization fails.
        """
        def auth_wrapper(self, *args, **kwargs):
            self.config.authenticate()
            self.config.save_config()
            if self.config.check_auth():
                try:
                    return func(self, *args, **kwargs)
                except SSLError:
                    click.secho(('SSL cert verification failed.\n  Try running '
                                 'gh configure --enterprise\n  and type '
                                 "'n' when asked whether to verify SSL certs."),
                                fg=self.config.clr_error)
                except MissingSchema:
                    click.secho('Invalid GitHub Enterprise url',
                                fg=self.config.clr_error)
                except AuthenticationFailed:
                    self.config.print_auth_error()
        return auth_wrapper

    def avatar(self, url, text_avatar):
        """Display the user's avatar from the specified url.

        This method requires PIL.

        :type url: str
        :param url: The user's avatar image.

        :type text_avatar: bool
        :param text_avatar: Determines whether to view the profile avatar
            in plain text (True) or in ansi (False).
            On Windows this value is always set to True due to lack of
            support of `img2txt` on Windows.

        :rtype: str
        :return: The avatar.
        """
        if platform.system() == 'Windows':
            text_avatar = True
        avatar = self.config.get_github_config_path(
            self.config.CONFIG_AVATAR)
        try:
            urllib.request.urlretrieve(url, avatar)
        except urllib.error.URLError:
            pass
        avatar_text = ''
        if os.path.exists(avatar):
            avatar_text = self.img2txt(avatar, ansi=(not text_avatar))
            avatar_text += '\n'
            os.remove(avatar)
        return avatar_text

    def avatar_setup(self, url, text_avatar):
        """Prepare to display the user's avatar from the specified url.

        This method requires PIL.

        :type url: str
        :param url: The user's avatar image.

        :type text_avatar: bool
        :param text_avatar: Determines whether to view the profile avatar
            in plain text (True) or in ansi (False).
            On Windows this value is always set to True due to lack of
            support of `img2txt` on Windows.

        :rtype: str
        :return: The avatar.
        """
        try:
            import PIL  # NOQA
            return self.avatar(url, text_avatar)
        except ImportError:
            avatar_text = click.style(('To view the avatar in your terminal, '
                                       'install the Python Image Library.\n'),
                                      fg=self.config.clr_message)
            return avatar_text

    def configure(self, enterprise):
        """Configure gitsome.

        Attempts to authenticate the user and to set up the user's news feed.

        If `gitsome` has not yet been configured, calling a `gh` command that
        requires authentication will automatically invoke the `configure`
        command.

        :type enterprise: bool
        :param enterprise: Determines whether to configure GitHub Enterprise.
        """
        self.config.authenticate(enterprise=enterprise, overwrite=True)
        self.config.prompt_news_feed()
        self.config.show_bash_completions_info()
        self.config.save_config()

    @authenticate
    def create_comment(self, user_repo_number, text):
        """Create a comment on the given issue.

        :type user_repo_number: str
        :param user_repo_number: The user/repo/issue_number.

        :type text: str
        :param text: The comment text.
        """
        try:
            user, repo, number = user_repo_number.split('/')
            int(number)  # Check for int
        except ValueError:
            click.secho(('Expected argument: user/repo/# and option -t '
                         '"comment".'),
                        fg=self.config.clr_error)
            return
        issue = self.config.api.issue(user, repo, number)
        issue_comment = issue.create_comment(text)
        if type(issue_comment) is not null.NullObject:
            click.secho('Created comment: ' + issue_comment.body,
                        fg=self.config.clr_message)
        else:
            click.secho('Error creating comment',
                        fg=self.config.clr_error)

    @authenticate
    def create_issue(self, user_repo, issue_title, issue_desc=''):
        """Create an issue.

        :type user_repo: str
        :param user_repo: The user/repo.

        :type issue_title: str
        :param issue_title: The issue title.

        :type issue_desc: str
        :param issue_desc: The issue body (optional).
        """
        try:
            user, repo_name = user_repo.split('/')
        except ValueError:
            click.secho('Expected argument: user/repo and option -t "title".',
                        fg=self.config.clr_error)
            return
        issue = self.config.api.create_issue(user,
                                             repo_name,
                                             issue_title,
                                             issue_desc)
        if type(issue) is not null.NullObject:
            body = self.text_utils.sanitize_if_none(issue.body)
            click.secho('Created issue: ' + issue.title + '\n' + body,
                        fg=self.config.clr_message)
        else:
            click.secho('Error creating issue.', fg=self.config.clr_error)

    @authenticate
    def create_repo(self, repo_name, repo_desc='', private=False):
        """Create a repo.

        :type repo_name: str
        :param repo_name: The repo name.

        :type repo_desc: str
        :param repo_desc: The repo description (optional).

        :type private: bool
        :param private: Determines whether the repo is private.  Default: False.
        """
        try:
            repo = self.config.api.create_repository(repo_name,
                                                     repo_desc,
                                                     private=private)
            desc = self.text_utils.sanitize_if_none(repo.description)
            click.secho(('Created repo: ' + repo.full_name + '\n' + desc),
                        fg=self.config.clr_message)
        except UnprocessableEntity as e:
            click.secho('Error creating repo: ' + str(e.msg),
                        fg=self.config.clr_error)

    @authenticate
    def emails(self):
        """List all the user's registered emails."""
        self.table.build_table_setup(self.config.api.emails(),
                                     self.formatter.format_email,
                                     limit=sys.maxsize,
                                     pager=False,
                                     build_urls=False)

    @authenticate
    def emojis(self, pager=False):
        """List all GitHub supported emojis.

        :type pager: bool
        :param pager: Determines whether to show the output in a pager,
            if available.
        """
        self.table.build_table_setup(self.config.api.emojis(),
                                     self.formatter.format_emoji,
                                     limit=sys.maxsize,
                                     pager=pager,
                                     build_urls=False)

    @authenticate
    def feed(self, user_or_repo='', private=False, pager=False):
        """List all activity for the given user or repo.

        If `user_or_repo` is not provided, uses the logged in user's news feed
        seen while visiting https://github.com.  If `user_or_repo` is provided,
        shows either the public or `[-p/--private]` feed activity of the user
        or repo.

        :type user_or_repo: str
        :param user_or_repo: The user or repo to list events for (optional).
            If no entry, defaults to the logged in user's feed.

        :type private: bool
        :param private: Determines whether to show the private events (True)
            or public events (False).

        :type pager: bool
        :param pager: Determines whether to show the output in a pager,
            if available.
        """
        click.secho('Listing events...', fg=self.config.clr_message)
        if user_or_repo == '':
            if self.config.user_feed is None:
                self.config.prompt_news_feed()
                self.config.save_config()
            if self.config.user_feed:
                items = self.trend_parser.parse(self.config.user_feed)
                self.table.build_table_setup_feed(
                    items,
                    self.formatter.format_feed_entry,
                    pager)
        else:
            if '/' in user_or_repo:
                user, repo = user_or_repo.split('/')
                repo = self.config.api.repository(user, repo)
                items = repo.events()
            else:
                public = False if private else True
                items = self.config.api.user(user_or_repo).events(public=public)
            self.table.build_table_setup(
                items,
                self.formatter.format_event,
                limit=sys.maxsize,
                pager=pager,
                build_urls=False)

    @authenticate
    def followers(self, user, pager=False):
        """List all followers and the total follower count.

        :type user: str
        :param user: The user login (optional).
            If None, returns the followers of the logged in user.

        :type pager: bool
        :param pager: Determines whether to show the output in a pager,
            if available.
        """
        if user is None:
            user = self.config.user_login
        self.table.build_table_setup_user(
            self.config.api.followers_of(user),
            self.formatter.format_user,
            limit=sys.maxsize,
            pager=pager)

    @authenticate
    def following(self, user, pager=False):
        """List all followed users and the total followed count.

        :type user: str
        :param user: The user login.
            If None, returns the followed users of the logged in user.

        :type pager: bool
        :param pager: Determines whether to show the output in a pager,
            if available.
        """
        if user is None:
            user = self.config.user_login
        self.table.build_table_setup_user(
            self.config.api.followed_by(user),
            self.formatter.format_user,
            limit=sys.maxsize,
            pager=pager)

    @authenticate
    def gitignore_template(self, language):
        """Output the gitignore template for the given language.

        :type language: str
        :param language: The language.
        """
        template = self.config.api.gitignore_template(language)
        if template:
            click.secho(template, fg=self.config.clr_message)
        else:
            click.secho(('Invalid case-sensitive template requested, run the '
                         'following command to see available templates:\n'
                         '    gh gitignore-templates'),
                        fg=self.config.clr_error)

    @authenticate
    def gitignore_templates(self, pager=False):
        """Output all supported gitignore templates.

        :type pager: bool
        :param pager: Determines whether to show the output in a pager,
            if available.
        """
        self.table.build_table_setup(
            self.config.api.gitignore_templates(),
            self.formatter.format_gitignore_template_name,
            limit=sys.maxsize,
            pager=pager,
            build_urls=False)
        click.secho(('  Run the following command to view or download a '
                     '.gitignore file:\n'
                     '    View:     gh gitignore Python\n'
                     '    Download: gh gitignore Python > .gitignore\n'),
                    fg=self.config.clr_message)

    @authenticate
    def issue(self, user_repo_number):
        """Output detailed information about the given issue.

        :type user_repo_number: str
        :param user_repo_number: The user/repo/issue_number.
        """
        try:
            user, repo, number = user_repo_number.split('/')
            int(number)  # Check for int
        except ValueError:
            click.secho('Expected argument: user/repo/#.',
                        fg=self.config.clr_error)
            return
        url = (self.base_url + user + '/' + repo + '/' +
               'issues/' + number)
        self.web_viewer.view_url(url)

    @authenticate
    def issues(self, issues_list, limit=1000, pager=False, sort=True):
        """List all issues.

        :type issues_list: list
        :param issues_list: A list of `github3` Issues.

        :type limit: int
        :param limit: The number of items to display.

        :type pager: bool
        :param pager: Determines whether to show the output in a pager,
            if available.

        :type sort: bool
        :param sort: Determines whether to sort the issues by:
            state, repo, created_at.
        """
        view_entries = []
        for current_issue in issues_list:
            url = self.formatter.format_issues_url_from_issue(current_issue)
            view_entries.append(
                ViewEntry(
                    current_issue,
                    url=url,
                    sort_key_primary=current_issue.state,
                    sort_key_secondary=current_issue.repository,
                    sort_key_tertiary=current_issue.created_at))
        if sort:
            view_entries = sorted(view_entries, reverse=False)
        self.table.build_table(view_entries,
                               limit,
                               pager,
                               self.formatter.format_issue)

    @authenticate
    def issues_setup(self, issue_filter='subscribed', issue_state='open',
                     limit=1000, pager=False):
        """Prepare to list all issues matching the filter.

        :type issue_filter: str
        :param issue_filter: 'assigned', 'created', 'mentioned',
            'subscribed' (default).

        :type issue_state: str
        :param issue_state: 'all', 'open' (default), 'closed'.

        :type limit: int
        :param limit: The number of items to display.

        :type pager: bool
        :param pager: Determines whether to show the output in a pager,
            if available.
        """
        self.issues(self.config.api.issues(issue_filter, issue_state),
                    limit,
                    pager)

    @authenticate
    def license(self, license_name):
        """Output the gitignore template for the given language.

        :type license_name: str
        :param license_name: The license name.
        """
        result = self.config.api.license(license_name)
        if type(result) is not null.NullObject:
            click.secho(result.body, fg=self.config.clr_message)
        else:
            click.secho(('  Invalid case-sensitive license requested, run the '
                         'following command to see available licenses:\n'
                         '    gh licenses'),
                        fg=self.config.clr_error)

    @authenticate
    def licenses(self):
        """Output the gitignore template for the given language."""
        self.table.build_table_setup(
            self.config.api.licenses(),
            self.formatter.format_license_name,
            limit=sys.maxsize,
            pager=False,
            build_urls=False)
        click.secho(('  Run the following command to view or download a '
                     'LICENSE file:\n'
                     '    gh license apache-2.0\n'
                     '    gh license apache-2.0 > LICENSE\n'),
                    fg=self.config.clr_message)

    @authenticate
    def notifications(self, limit=1000, pager=False):
        """List all notifications.

        :type limit: int
        :param limit: The number of items to display.

        :type pager: bool
        :param pager: Determines whether to show the output in a pager,
            if available.
        """
        view_entries = []
        for thread in self.config.api.notifications(all=True,
                                                    participating=False):
            url = self.formatter.format_issues_url_from_thread(thread)
            view_entries.append(ViewEntry(thread, url=url))
        self.table.build_table(view_entries,
                               limit,
                               pager,
                               self.formatter.format_thread)

    @authenticate
    def octocat(self, say):
        """Output an Easter egg or the given message from Octocat.

        :type say: str
        :param say: What Octocat should say.
                If say is None, octocat speaks an Easter egg.
        """
        output = str(self.config.api.octocat(say))
        output = output.replace('\\n', '\n')
        click.secho(output, fg=self.config.clr_message)

    @authenticate
    def pull_requests(self, limit=1000, pager=False):
        """List all pull requests.

        :type limit: int
        :param limit: The number of items to display.

        :type pager: bool
        :param pager: Determines whether to show the output in a pager,
            if available.
        """
        issues_list = []
        repositories = self.config.api.repositories()
        for repository in repositories:
            repo_pulls = repository.pull_requests()
            for repo_pull in repo_pulls:
                url = self.formatter.format_issues_url_from_issue(repo_pull)
                user, repo, issues, number = url.split('/')
                repo_pull = self.config.api.pull_request(user, repo, number)
                issues_list.append(repo_pull)
        self.issues(issues_list, limit, pager)

    @authenticate
    def rate_limit(self):
        """Output the rate limit.  Not available for GitHub Enterprise.

        Logged in users can make 5000 requests per hour.
        See: https://developer.github.com/v3/#rate-limiting
        """
        click.secho('Rate limit: ' + str(self.config.api.ratelimit_remaining),
                    fg=self.config.clr_message)

    @authenticate
    def repositories(self, repos, limit=1000, pager=False,
                     repo_filter='', print_output=True, sort=True):
        """List all repos matching the given filter.

        :type repos: list
        :param repos: A list of `github3` Repository.

        :type limit: int
        :param limit: The number of items to display.

        :type pager: bool
        :param pager: Determines whether to show the output in a pager,
            if available.

        :type repo_filter: str
        :param repo_filter:  The filter for repo names.
            Only repos matching the filter will be returned.
            If None, outputs all repos.

        :type print_output: bool
        :param print_output: Determines whether to print the output
                (True) or return the output as a string (False).

        :type sort: bool
        :param sort: Determines whether to sort the issues by:
            state, repo, created_at.

        :rtype: str
        :return: The output if print_output is True
            else, returns None.
        """
        view_entries = []
        for repo in repos:
            url = repo.clone_url
            if (repo.full_name is not None and
                    repo_filter in repo.full_name.lower()) or \
               (repo.description is not None and
                    repo_filter in repo.description.lower()):
                view_entries.append(
                    ViewEntry(repo,
                              url=url,
                              sort_key_primary=repo.stargazers_count))
        if sort:
            view_entries = sorted(view_entries, reverse=True)
        return self.table.build_table(view_entries,
                                      limit,
                                      pager,
                                      self.formatter.format_repo,
                                      print_output=print_output)

    @authenticate
    def repositories_setup(self, repo_filter, limit=1000, pager=False):
        """Prepare to list all repos matching the given filter.

        :type repo_filter: str
        :param repo_filter:  The filter for repo names.
            Only repos matching the filter will be returned.
            If None, outputs all repos retrieved by the GitHub API.

        :type limit: int
        :param limit: The number of items to display.

        :type pager: bool
        :param pager: Determines whether to show the output in a pager,
            if available.
        """
        self.repositories(self.config.api.repositories(),
                          limit,
                          pager,
                          repo_filter)

    @authenticate
    def repository(self, user_repo):
        """Output detailed information about the given repo.

        :type user_repo: str
        :param user_repo: The user/repo.
        """
        try:
            user, repo = user_repo.split('/')
        except ValueError:
            click.secho('Expected argument: user/repo.',
                        fg=self.config.clr_error)
            return
        self.web_viewer.view_url(self.base_url + user_repo)

    @authenticate
    def search_issues(self, query, limit=1000, pager=False):
        """Search for all issues matching the given query.

        :type query: str
        :param query: The search query.

        :type limit: int
        :param limit: The number of items to display.

        :type pager: bool
        :param pager: Determines whether to show the output in a pager,
            if available.
        """
        click.secho('Searching for all matching issues on GitHub...',
                    fg=self.config.clr_message)
        results = self.config.api.search_issues(query)
        issues_list = []
        for result in results:
            issues_list.append(result.issue)
        self.issues(issues_list, limit, pager, sort=False)

    @authenticate
    def search_repositories(self, query, sort, limit=1000, pager=False):
        """Search for all repos matching the given query.

        :type query: str
        :param query: The search query.

        :type sort: str
        :param sort: Optional: 'stars', 'forks', 'updated'.
            If not specified, sorting is done by query best match.

        :type limit: int
        :param limit: The number of items to display.

        :type pager: bool
        :param pager: Determines whether to show the output in a pager,
            if available.
        """
        click.secho('Searching for all matching repos on GitHub...',
                    fg=self.config.clr_message)
        results = self.config.api.search_repositories(query, sort)
        repos = []
        for result in results:
            repos.append(result.repository)
        self.repositories(repos, limit, pager, sort=False)

    @authenticate
    def starred(self, repo_filter, limit=1000, pager=False):
        """Output starred repos.

        :type repo_filter: str
        :param repo_filter:  The filter for repo names.
            Only repos matching the filter will be returned.
            If None, outputs all starred repos.

        :type limit: int
        :param limit: The number of items to display.

        :type pager: bool
        :param pager: Determines whether to show the output in a pager,
            if available.
        """
        self.repositories(self.config.api.starred(),
                          limit,
                          pager,
                          repo_filter.lower())

    def trending(self, language, weekly, monthly,
                 devs=False, browser=False, pager=False):
        """List trending repos for the given language.

        :type language: str
        :param language: The language (optional).
            If blank, shows 'Overall'.

        :type weekly: bool
        :param weekly: Determines whether to show the weekly rankings.
            Daily is the default.

        :type monthly: bool
        :param monthly: Determines whether to show the monthly rankings.
            Daily is the default.
            If both `monthly` and `weekly` are set, `monthly` takes precedence.

        :type devs: bool
        :param devs: determines whether to display the trending
                devs or repos.  Only valid with the -b/--browser option.

        :type browser: bool
        :param browser: Determines whether to view the profile
                in a browser, or in the terminal.

        :type pager: bool
        :param pager: Determines whether to show the output in a pager,
            if available.
        """
        language = language.lower()
        if language in language_rss_map:
            language = language_rss_map[language]
        if monthly:
            period = 'monthly'
            url_param = '?since=monthly'
        elif weekly:
            period = 'weekly'
            url_param = '?since=weekly'
        else:
            period = 'daily'
            url_param = ''
        if browser:
            webbrowser.open(
                ('https://github.com/trending' +
                 ('/developers' if devs else '') +
                 ('/' + language if language is not 'overall' else '') +
                 url_param))
        else:
            click.secho(
                'Listing {p} trending {l} repos...'.format(l=language,
                                                           p=period),
                fg=self.config.clr_message)
            url = ('http://github-trends.ryotarai.info/rss/github_trends_' +
                   language + '_')
            url += period + '.rss'
            items = self.trend_parser.parse(url)
            self.table.build_table_setup_trending(
                items.entries,
                self.formatter.format_trending_entry,
                limit=sys.maxsize,
                pager=pager)

    @authenticate
    def user(self, user_id, browser=False, text_avatar=False,
             limit=1000, pager=False):
        """List information about the logged in user.

        :type user_id: str
        :param user_id: The user id/login.
            If None, returns followers of the logged in user.

        :type browser: bool
        :param browser: Determines whether to view the profile
            in a browser, or in the terminal.

        :type text_avatar: bool
        :param text_avatar: Determines whether to view the profile
            avatar in plain text instead of ansi (default).
            On Windows this value is always set to True due to lack of
            support of `img2txt` on Windows.

        :type limit: int
        :param limit: The number of items to display.

        :type pager: bool
        :param pager: Determines whether to show the output in a pager,
            if available.
        """
        if browser:
            webbrowser.open(self.base_url + user_id)
        else:
            user = self.config.api.user(user_id)
            if type(user) is null.NullObject:
                click.secho('Invalid user.', fg=self.config.clr_error)
                return
            output = ''
            output += click.style(self.avatar_setup(user.avatar_url,
                                                    text_avatar))
            output += click.style(user.login + '\n', fg=self.config.clr_primary)
            if user.company is not None:
                output += click.style(user.company + '\n',
                                      fg=self.config.clr_secondary)
            if user.location is not None:
                output += click.style(user.location + '\n',
                                      fg=self.config.clr_secondary)
            if user.email is not None:
                output += click.style(user.email + '\n',
                                      fg=self.config.clr_secondary)
            if user.type == 'Organization':
                output += click.style('Organization\n\n',
                                      fg=self.config.clr_tertiary)
            else:
                output += click.style(
                    'Followers: ' + str(user.followers_count) + ' | ',
                    fg=self.config.clr_tertiary)
                output += click.style(
                    'Following: ' + str(user.following_count) + '\n\n',
                    fg=self.config.clr_tertiary)
            output += self.repositories(self.config.api.repositories(user_id),
                                        limit,
                                        pager,
                                        print_output=False)
            if pager:
                color = None
                if platform.system() == 'Windows':
                    color = True
                click.echo_via_pager(output, color)
            else:
                click.secho(output)

    @authenticate
    def user_me(self, browser, text_avatar, limit=1000, pager=False):
        """List information about the logged in user.

        :type browser: bool
        :param browser: Determines whether to view the profile
                in a browser, or in the terminal.

        :type text_avatar: bool
        :param text_avatar: Determines whether to view the profile
            avatar in plain text.
            On Windows this value is always set to True due to lack of
            support of `img2txt` on Windows.

        :type limit: int
        :param limit: The number of items to display.

        :type pager: bool
        :param pager: Determines whether to show the output in a pager,
            if available.
        """
        self.user(self.config.user_login, browser, text_avatar, limit, pager)

    @authenticate
    def view(self, index, view_in_browser=False):
        """View the given index in a browser.

        Load urls from ~/.gitsomeconfigurl and stores them in self.config.urls.
        Open a browser with the url based on the given index.

        :type index: int
        :param index: Determines the index to view.

        :type browser: bool
        :param browser: Determines whether to view the profile
            in a browser, or in the terminal.
        """
        self.config.urls = self.config.load_urls(view_in_browser)
        url = self.config.urls[index-1]
        click.secho('Viewing ' + url + '...', fg=self.config.clr_message)
        if view_in_browser:
            webbrowser.open(self.add_base_url(url))
        else:
            if 'issues/' in url:
                url = url.replace('issues/', '')
                self.issue(url)
            elif len(url.split('/')) == 2:
                self.repository(url)
            else:
                self.web_viewer.view_url(self.add_base_url(url))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2015 Donne Martin. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import unicode_literals
from __future__ import print_function

from .githubcli import GitHubCli


def cli():
    """Create and call GitHubCli."""
    github = GitHubCli()
    github.cli()


if __name__ == "__main__":
    cli()



"""The module that holds much of the metadata about github3.py."""
__package_name__ = 'github3.py'
__title__ = 'github3'
__author__ = 'Ian Cordasco'
__author_email__ = 'graffatcolmingov@gmail.com'
__license__ = 'Modified BSD'
__copyright__ = 'Copyright 2012-2016 Ian Cordasco'
__version__ = '1.0.0a4'
__version_info__ = tuple(int(i) for i in __version__.split('.') if i.isdigit())
__url__ = 'http://github3.readthedocs.org'

__all__ = (
    '__package_name__', '__title__', '__author__', '__author_email__',
    '__license__', '__copyright__', '__version__', '__version_info__',
    '__url__',
)



# -*- coding: utf-8 -*-
import collections
import functools

from requests.compat import urlparse, urlencode

from . import exceptions
from . import models


class GitHubIterator(models.GitHubCore, collections.Iterator):
    """The :class:`GitHubIterator` class powers all of the iter_* methods."""
    def __init__(self, count, url, cls, session, params=None, etag=None,
                 headers=None):
        models.GitHubCore.__init__(self, {}, session)
        #: Original number of items requested
        self.original = count
        #: Number of items left in the iterator
        self.count = count
        #: URL the class used to make it's first GET
        self.url = url
        #: Last URL that was requested
        self.last_url = None
        self._api = self.url
        #: Class for constructing an item to return
        self.cls = cls
        #: Parameters of the query string
        self.params = params or {}
        self._remove_none(self.params)
        # We do not set this from the parameter sent. We want this to
        # represent the ETag header returned by GitHub no matter what.
        # If this is not None, then it won't be set from the response and
        # that's not what we want.
        #: The ETag Header value returned by GitHub
        self.etag = None
        #: Headers generated for the GET request
        self.headers = headers or {}
        #: The last response seen
        self.last_response = None
        #: Last status code received
        self.last_status = 0

        if etag:
            self.headers.update({'If-None-Match': etag})

        self.path = urlparse(self.url).path

    def _repr(self):
        return '<GitHubIterator [{0}, {1}]>'.format(self.count, self.path)

    def __iter__(self):
        self.last_url, params = self.url, self.params
        headers = self.headers

        if 0 < self.count <= 100 and self.count != -1:
            params['per_page'] = self.count

        if 'per_page' not in params and self.count == -1:
            params['per_page'] = 100

        cls = self.cls
        if issubclass(self.cls, models.GitHubCore):
            cls = functools.partial(self.cls, session=self)

        while (self.count == -1 or self.count > 0) and self.last_url:
            response = self._get(self.last_url, params=params,
                                 headers=headers)
            self.last_response = response
            self.last_status = response.status_code
            if params:
                params = None  # rel_next already has the params

            if not self.etag and response.headers.get('ETag'):
                self.etag = response.headers.get('ETag')

            json = self._get_json(response)

            if json is None:
                break

            # languages returns a single dict. We want the items.
            if isinstance(json, dict):
                if issubclass(self.cls, models.GitHubCore):
                    raise exceptions.UnprocessableResponseBody(
                        "GitHub's API returned a body that could not be"
                        " handled", json
                    )
                if json.get('ETag'):
                    del json['ETag']
                if json.get('Last-Modified'):
                    del json['Last-Modified']
                json = json.items()

            for i in json:
                yield cls(i)
                self.count -= 1 if self.count > 0 else 0
                if self.count == 0:
                    break

            rel_next = response.links.get('next', {})
            self.last_url = rel_next.get('url', '')

    def __next__(self):
        if not hasattr(self, '__i__'):
            self.__i__ = self.__iter__()
        return next(self.__i__)

    def _get_json(self, response):
        return self._json(response, 200)

    def refresh(self, conditional=False):
        self.count = self.original
        if conditional:
            self.headers['If-None-Match'] = self.etag
        self.etag = None
        self.__i__ = self.__iter__()
        return self

    def next(self):
        return self.__next__()


class SearchIterator(GitHubIterator):

    """This is a special-cased class for returning iterable search results.

    It inherits from :class:`GitHubIterator <github3.structs.GitHubIterator>`.
    All members and methods documented here are unique to instances of this
    class. For other members and methods, check its parent class.

    """

    def __init__(self, count, url, cls, session, params=None, etag=None,
                 headers=None):
        super(SearchIterator, self).__init__(count, url, cls, session, params,
                                             etag, headers)
        #: Total count returned by GitHub
        self.total_count = 0
        #: Items array returned in the last request
        self.items = []

    def _repr(self):
        return '<SearchIterator [{0}, {1}?{2}]>'.format(self.count, self.path,
                                                        urlencode(self.params))

    def _get_json(self, response):
        json = self._json(response, 200)
        # I'm not sure if another page will retain the total_count attribute,
        # so if it's not in the response, just set it back to what it used to
        # be
        self.total_count = json.get('total_count', self.total_count)
        self.items = json.get('items', [])
        # If we return None then it will short-circuit the while loop.
        return json.get('items')



# -*- coding: utf-8 -*-

# Copyright 2015 Donne Martin. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

formatted_events = u'\x1b[35m  1.   \x1b[0mdonnemartin \x1b[0m\x1b[32mcommented on commit \x1b[0m\x1b[36mAAA23e2\x1b[0m\x1b[32m at \x1b[0m\x1b[36muser1/repo1\x1b[0m\x1b[33m (just now)\x1b[0m\n         \x1b[36m\x1b[0mfoo\x1b[0m\n\x1b[35m  2.   \x1b[0mdonnemartin \x1b[0m\x1b[32mcreated\x1b[0m\x1b[32m branch\x1b[0m\x1b[36m master\x1b[0m\x1b[32m at \x1b[0m\x1b[36muser1/repo1\x1b[0m\x1b[33m (just now)\x1b[0m\n         \x1b[36m\x1b[0m\x1b[0m\n\x1b[35m  3.   \x1b[0mdonnemartin \x1b[0m\x1b[32mfollowed \x1b[0m\x1b[36muser1/repo1\x1b[0m\x1b[33m (just now)\x1b[0m\n\x1b[35m  4.   \x1b[0mdonnemartin \x1b[0m\x1b[32mforked\x1b[0m\x1b[36m user1/repo1\x1b[0m\x1b[33m (just now)\x1b[0m\n\x1b[35m  5.   \x1b[0mdonnemartin \x1b[0m\x1b[32mcommented on \x1b[0m\x1b[36muser1/repo1#1\x1b[0m\x1b[33m (just now)\x1b[0m\n         \x1b[36m\x1b[0mfoo\x1b[0m\n           \x1b[36m\x1b[0mfoo\x1b[0m\n\x1b[35m  6.   \x1b[0mdonnemartin \x1b[0m\x1b[32mclosed issue \x1b[0m\x1b[36muser1/repo1#1\x1b[0m\x1b[33m (just now)\x1b[0m\n         \x1b[36m\x1b[0mfoo\x1b[0m\n\x1b[35m  7.   \x1b[0mdonnemartin \x1b[0m\x1b[32mclosed pull request \x1b[0m\x1b[36muser1/repo1#1\x1b[0m\x1b[33m (just now)\x1b[0m\n         \x1b[36m\x1b[0mfoo\x1b[0m\n\x1b[35m  8.   \x1b[0mdonnemartin \x1b[0m\x1b[32mpushed to\x1b[0m\x1b[36m master\x1b[0m\x1b[32m at \x1b[0m\x1b[36muser1/repo1\x1b[0m\x1b[33m (just now)\x1b[0m\n         \x1b[36m5ee4d1b: \x1b[0m\x1b[0mFix GitHubCli class docstring\x1b[0m\n         \x1b[36mfc2309b: \x1b[0m\x1b[0mUpdate gh configure docstring\x1b[0m\n         \x1b[36mdde19b7: \x1b[0m\x1b[0mUpdate gh create-comment docstring\x1b[0m\n\x1b[35m  9.   \x1b[0mdonnemartin \x1b[0m\x1b[32mreleased \x1b[0m\x1b[36m0.5.0 \x1b[0m\x1b[32mat \x1b[0m\x1b[36muser1/repo1\x1b[0m\x1b[33m (just now)\x1b[0m\n\x1b[0m'



#!/usr/bin/env python
# coding=utf-8

"""The gitsome installer."""
from __future__ import print_function, unicode_literals
import os
import sys
try:
    from setuptools import setup, find_packages
    from setuptools.command.sdist import sdist
    from setuptools.command.install import install
    from setuptools.command.develop import develop
    HAVE_SETUPTOOLS = True
except ImportError:
    from distutils.core import setup
    from distutils.command.sdist import sdist as sdist
    from distutils.command.install import install as install
    HAVE_SETUPTOOLS = False
from gitsome.__init__ import __version__ as VERSION


TABLES = ['xonsh/lexer_table.py', 'xonsh/parser_table.py']


def clean_tables():
    for f in TABLES:
        if os.path.isfile(f):
            os.remove(f)
            print('Remove ' + f)


def build_tables():
    print('Building lexer and parser tables.')
    sys.path.insert(0, os.path.dirname(__file__))
    from xonsh.parser import Parser
    Parser(lexer_table='lexer_table', yacc_table='parser_table',
           outputdir='xonsh')
    sys.path.pop(0)


class xinstall(install):
    def run(self):
        clean_tables()
        build_tables()
        install.run(self)


class xsdist(sdist):
    def make_release_tree(self, basedir, files):
        clean_tables()
        build_tables()
        sdist.make_release_tree(self, basedir, files)


if HAVE_SETUPTOOLS:
    class xdevelop(develop):
        def run(self):
            clean_tables()
            build_tables()
            develop.run(self)


def main():
    if sys.version_info < (3, 4):
        print('gitsome requires at least Python 3.4.')
        sys.exit(1)
    try:
        if '--name' not in sys.argv:
            print(logo)
    except UnicodeEncodeError:
        pass
    skw = dict(
        name='gitsome',
        description='A Supercharged Git/Shell Autocompleter with GitHub Integration.',  # NOQA
        license='Apache License 2.0',
        version=VERSION,
        author='Donne Martin',
        maintainer='Donne Martin',
        author_email='donne.martin@gmail.com',
        url='https://github.com/donnemartin/gitsome',
        platforms='Cross Platform',
        classifiers=[
            'Intended Audience :: Developers',
            'Intended Audience :: System Administrators',
            'License :: OSI Approved :: Apache Software License',
            'Natural Language :: English',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Topic :: Software Development',
            'Topic :: Software Development :: Libraries :: Python Modules',
        ],
        packages=find_packages(),
        scripts=['scripts/xonsh',
                 'scripts/xonsh.bat',
                 'scripts/gitsome',
                 'scripts/gitsome.bat'],
        cmdclass={'install': xinstall, 'sdist': xsdist},
        )
    if HAVE_SETUPTOOLS:
        skw['setup_requires'] = ['ply']
        skw['install_requires'] = [
            'numpydoc>=0.5,<1.0',
            'ply>=3.4,<4.0',
            'prompt_toolkit>=2.0.0,<2.1.0',
            'requests>=2.8.1,<3.0.0',
            'colorama>=0.3.3,<1.0.0',
            'click>=5.1,<7.0',
            'pygments>=2.0.2,<3.0.0',
            'feedparser>=5.2.1,<6.0.0',
            'pytz>=2016.3,<2017.0',
            'docopt>=0.6.2,<1.0.0',
            'uritemplate.py>=1.0.0,<4.0.0',
        ],
        skw['entry_points'] = {
            'pygments.lexers': ['gitsome = xonsh.pyghooks:XonshLexer',
                                'gitsomecon = xonsh.pyghooks:XonshConsoleLexer',
                                ],
            'console_scripts': ['gh = gitsome.main_cli:cli',
                                ],
            }
        skw['cmdclass']['develop'] = xdevelop
    setup(**skw)


logo = ''


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""History object for use with prompt_toolkit."""
import builtins

import prompt_toolkit.history


class PromptToolkitHistory(prompt_toolkit.history.History):
    """History class that implements the prompt-toolkit history interface
    with the xonsh backend.
    """

    def __init__(self, load_prev=True, *args, **kwargs):
        """Initialize history object."""
        super().__init__()
        self.load_prev = load_prev

    def store_string(self, entry):
        pass

    def load_history_strings(self):
        """Loads synchronous history strings"""
        if not self.load_prev:
            return
        hist = builtins.__xonsh__.history
        if hist is None:
            return
        for cmd in hist.all_items(newest_first=True):
            line = cmd["inp"].rstrip()
            strs = self.get_strings()
            if len(strs) == 0 or line != strs[-1]:
                yield line

    def __getitem__(self, index):
        return self.get_strings()[index]

    def __len__(self):
        return len(self.get_strings())

    def __iter__(self):
        return iter(self.get_strings())


def _cust_history_matches(self, i):
    """Custom history search method for prompt_toolkit that matches previous
    commands anywhere on a line, not just at the start.

    This gets monkeypatched into the prompt_toolkit prompter if
    ``XONSH_HISTORY_MATCH_ANYWHERE=True``"""
    return (
        self.history_search_text is None
        or self.history_search_text in self._working_lines[i]
    )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from ..models import GitHubCore
from ..users import User


class IssueEvent(GitHubCore):
    """The :class:`IssueEvent <IssueEvent>` object. This specifically deals
    with events described in the
    `Issues\>Events <http://developer.github.com/v3/issues/events>`_ section of
    the GitHub API.

    Two event instances can be checked like so::

        e1 == e2
        e1 != e2

    And is equivalent to::

        e1.commit_id == e2.commit_id
        e1.commit_id != e2.commit_id

    """
    def _update_attributes(self, event):
        # The type of event:
        #   ('closed', 'reopened', 'subscribed', 'merged', 'referenced',
        #    'mentioned', 'assigned')
        #: The type of event, e.g., closed
        self.event = event.get('event')
        #: SHA of the commit.
        self.commit_id = event.get('commit_id')
        self._api = event.get('url', '')

        #: :class:`Issue <github3.issues.Issue>` where this comment was made.
        self.issue = event.get('issue')
        if self.issue:
            from .issue import Issue
            self.issue = Issue(self.issue, self)

        #: :class:`User <github3.users.User>` who caused this event.
        self.actor = event.get('actor')
        if self.actor:
            self.actor = User(self.actor, self)

        #: :class:`User <github3.users.User>` that generated the event.
        self.actor = event.get('actor')
        if self.actor:
            self.actor = User(self.actor, self)

        #: Number of comments
        self.comments = event.get('comments', 0)

        #: datetime object representing when the event was created.
        self.created_at = self._strptime(event.get('created_at'))

        #: Dictionary of links for the pull request
        self.pull_request = event.get('pull_request', {})

        #: Dictionary containing label details
        self.label = event.get('label', {})

        #: The integer ID of the event
        self.id = event.get('id')

        #: :class:`User <github3.users.User>` that is assigned
        self.assignee = event.get('assignee')
        if self.assignee:
            self.assignee = User(self.assignee, self)

        #: Dictionary containing milestone details
        self.milestone = event.get('milestone', {})

        #: Dictionary containing to and from attributes
        self.rename = event.get('rename', {})

        self._uniq = self.commit_id

    def _repr(self):
        return '<Issue Event [{0} by {1}]>'.format(
            self.event, self.actor
            )



# must come before ptk / pygments imports
from xonsh.lazyasd import load_module_in_background

load_module_in_background(
    "pkg_resources",
    debug="XONSH_DEBUG",
    replacements={"pygments.plugin": "pkg_resources"},
)



# -*- coding: utf-8 -*-

# Copyright 2015 Donne Martin. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import unicode_literals
from __future__ import print_function

from datetime import datetime
import mock
import pytz

from gitsome.lib.github3 import null
from gitsome.lib.github3.exceptions import UnprocessableEntity


class MockEmail(object):

    def __init__(self, email, primary=False, verified=False):
        self.email = email
        self.primary = primary
        self.verified = verified


class MockUser(object):

    def __init__(self, login, user_type='User'):
        self.login = login
        self.repositories = {}
        self.emails = []
        self.avatar_url = 'https://www.github.com'
        self.company = 'bigcorp'
        self.location = 'interwebz'
        self.email = login + '@foo.com'
        self.type = user_type
        self.followers_count = 0
        self.following_count = 0
        self.feed_events = []

    def events(self, public):
        feed_events = []
        feed_events.append(MockEvent(
            event_type='CommitCommentEvent',
            payload={
                'comment': MockRepoComment(),
            }))
        feed_events.append(MockEvent(
            event_type='CreateEvent',
            payload={
                'ref_type': 'branch',
                'ref': 'master',
                'description': None,
            }))
        feed_events.append(MockEvent(
            event_type='FollowEvent'))
        feed_events.append(MockEvent(
            event_type='ForkEvent'))
        feed_events.append(MockEvent(
            event_type='IssueCommentEvent',
            payload={
                'comment': MockIssueComment('foo'),
                'issue': MockIssue('1', MockRepo(self, 'repo1'), 'foo'),
            }))
        feed_events.append(MockEvent(
            event_type='IssuesEvent',
            payload={
                'action': 'closed',
                'issue': MockIssue('1', MockRepo(self, 'repo1'), 'foo'),
            }))
        feed_events.append(MockEvent(
            event_type='PullRequestEvent',
            payload={
                'action': 'closed',
                'pull_request': MockIssue('1', MockRepo(self, 'repo1'), 'foo'),
            }))
        feed_events.append(MockEvent(
            event_type='PushEvent',
            payload={
                'ref': 'refs/heads/master',
                'commits': [{'url': 'https://api.github.com/repos/donnemartin/gitsome/commits/5ee4d1b20ee7cb16cd5be19b103301541a41003f', 'message': 'Fix GitHubCli class docstring', 'distinct': True, 'author': {'email': 'donne.martin@gmail.com', 'name': 'Donne Martin'}, 'sha': '5ee4d1b20ee7cb16cd5be19b103301541a41003f'}, {'url': 'https://api.github.com/repos/donnemartin/gitsome/commits/fc2309b645313646a3792eca9e0e9168cf25b267', 'message': 'Update gh configure docstring', 'distinct': True, 'author': {'email': 'donne.martin@gmail.com', 'name': 'Donne Martin'}, 'sha': 'fc2309b645313646a3792eca9e0e9168cf25b267'}, {'url': 'https://api.github.com/repos/donnemartin/gitsome/commits/dde19b7685ad7a07872fea1b4dc8019585322fdb', 'message': 'Update gh create-comment docstring', 'distinct': True, 'author': {'email': 'donne.martin@gmail.com', 'name': 'Donne Martin'}, 'sha': 'dde19b7685ad7a07872fea1b4dc8019585322fdb'}]  # NOQA
            }))
        mock_release_tag = mock.Mock()
        mock_release_tag.tag_name = '0.5.0'
        feed_events.append(MockEvent(
            event_type='ReleaseEvent',
            payload={
                'release': mock_release_tag,
            }))
        return feed_events

    def raise_mock_unprocessableentity(self):
        response = mock.Mock()
        response.json = lambda: exec('raise(Exception())')
        response.content = 'foobar'
        raise UnprocessableEntity(response)

    def create_repo(self, name, desc='', private=False):
        if name in self.repositories:
            self.raise_mock_unprocessableentity()
        repo = MockRepo(self, name, desc, private)
        self.repositories.update({repo.full_name: repo})
        return repo


class MockRepo(object):

    def __init__(self, user, full_name, description='', private=False):
        self.user = user
        self.full_name = full_name
        self.description = description
        self.private = private
        self.issues = {}
        self.clone_url = 'https://github.com/octocat/spoon-knife'
        self.stargazers_count = 1
        self.forks_count = 1
        self.language = ''
        self.updated_at = ''
        self.repository = 'foobar'

    def __lt__(self, other):
        return self.full_name < other.full_name

    def gen_key(self):
        return len(self.issues) + 1

    def create_issue(self, issue_title, issue_desc=''):
        number = self.gen_key()
        issue = MockIssue(number, self, issue_title, issue_desc)
        self.issues.update({number: issue})
        return issue

    def pull_requests(self):
        return list(self.issues.values())


class MockIssue(object):

    def __init__(self, number, repository, title, body=''):
        self.number = number
        self.repository = (repository.user.login, repository.full_name)
        self.title = title
        self.body = body
        self.state = 'open'
        self.comments_count = 1
        self.assignee = 'user1'
        self.user = 'user2'
        self.created_at = ''
        self.comments = []
        self.issue = 'foobar'

    def create_comment(self, body):
        issue_comment = MockIssueComment(body)
        self.comments.append(issue_comment)
        return issue_comment


class MockIssueComment(object):

    def __init__(self, body):
        self.body = body


class MockRepoComment(object):

    def __init__(self):
        self.commit_id = 'AAA23e2c6cb6997d25cfe61673aea6d701e9bZZZ'
        self.body = 'foo'


class MockLicense(object):

    def __init__(self, key, name):
        self.key = key
        self.name = name


class MockThread(object):

    def __init__(self, thread_type, title, unread):
        self.subject = {
            'title': title,
            'type': thread_type,
            'url': 'https://api.github.com/repos/foo/bar/pulls/1',
        }
        self.unread = unread
        self.updated_at = ''


class MockEvent(object):

    def __init__(self, event_type, payload=''):
        self.id = 1
        self.created_at = datetime.now(pytz.utc)
        self.actor = 'donnemartin'
        self.org = 'org'
        self.type = event_type
        self.payload = payload
        self.repo = ('user1', 'repo1')
        self.public = True


class MockGitHubApi(object):

    def __init__(self):
        self.users = {}
        self.current_user = 'user1'
        self.ratelimit_remaining = 5000
        self._generate_mock_data()

    def _generate_mock_data(self):
        user1 = MockUser(self.current_user, 'User')
        user1_repo1 = user1.create_repo('repo1')
        user1_repo1.create_issue('title1', 'body1')
        user1_repo1.create_issue('title2', 'body2')
        user1_repo1.create_issue('title3', 'body3')
        user1.emails.extend([
            MockEmail('foo@baz.com', True, False),
            MockEmail('bar@baz.com', False, True),
        ])
        user2 = MockUser('user2', 'Organization')
        self.users.update({
            user1.login: user1,
            user2.login: user2,
        })

    def create_issue(self, user_login, repo_name, issue_title, issue_desc):
        try:
            user = self.users[user_login]
            repo = user.repositories[repo_name]
            issue = repo.create_issue(issue_title, issue_desc)
            return issue
        except KeyError:
            return null.NullObject('Issue')

    def create_repository(self, repo_name, repo_desc='', private=False):
        user = self.users[self.current_user]
        return user.create_repo(repo_name, repo_desc, private)

    def emails(self):
        user = self.users[self.current_user]
        return user.emails

    def emojis(self, pager=False):
        return [
            'dolls',
            'palm_tree',
            'uk',
            '100',
            'baby_chick',
        ]

    def followers_of(self, user_login):
        return [
            MockUser('foo1'),
            MockUser('foo2'),
            MockUser('foo3'),
        ]

    def followed_by(self, user_login):
        return self.followers_of(user_login)

    def gitignore_template(self, language):
        if language == 'valid_language':
            return 'template'
        else:
            return ''

    def gitignore_templates(self):
        return [
            'Actionscript',
            'Ada',
            'Agda',
            'Android',
            'AppEngine',
        ]

    def issue(self, user_login, repo_name, number):
        try:
            user = self.users[user_login]
            repo = user.repositories[repo_name]
            return repo.issues[int(number)]
        except KeyError:
            return null.NullObject('Issue')

    def issues(self, issue_filter='subscribed', issue_state='open'):
        user = self.users[self.current_user]
        repo = user.repositories['repo1']
        issues_dict = repo.issues
        issues = list(issues_dict.values())
        return issues

    def license(self, license):
        if license == 'valid_license':
            template = mock.Mock()
            template.body = 'template'
            return template
        else:
            return null.NullObject('License')

    def licenses(self):
        return [
            MockLicense('mit', '(MIT License)'),
            MockLicense('gpl-2.0', '(GNU General Public License v2.0)'),
            MockLicense('bsd-2-clause', '(BSD 2-clause "Simplified" License)'),
            MockLicense('isc', '(ISC License)'),
            MockLicense('epl-1.0', '(Eclipse Public License 1.0)'),
        ]

    def notifications(self, all=True, participating=False):
        return [
            MockThread('type1', 'title1', True),
            MockThread('type2', 'title2', False),
            MockThread('type3', 'title3', True),
        ]

    def octocat(self, say):
        return say

    def pull_request(self, owner, repository, number):
        pull_requests = self.issues()
        return pull_requests[0]

    def search_issues(self, query):
        return self.issues()

    def search_repositories(self, query, sort):
        return self.repositories()

    def repositories(self, user_id=None):
        if user_id is None:
            user_id = self.current_user
        user = self.users[user_id]
        repos = list(user.repositories.values())
        repos_sorted = sorted(repos)
        return repos_sorted

    def user(self, user_id):
        try:
            return self.users[user_id]
        except KeyError:
            return null.NullObject('User')



# -*- coding: utf-8 -*-

# Copyright 2015 Donne Martin. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

META_LOOKUP_GIT = {
    'git': 'The stupid content tracker.',
    '--version': 'Prints the Git suite version that the git program came from.',
    '--help': 'Prints the synopsis and a list of the most commonly used commands.',
    '-C': 'Run as if git was started in <path> instead of the current working directory.',
    '-c': 'Pass a configuration parameter to the command.',
    '--exec-path': 'Path to wherever your core Git programs are installed.',
    '--html-path': "Print the path, without trailing slash, where Git's HTML documentation is installed and exit.",
    '--man-path': 'Print the manpath (see man(1)) for the man pages for this version of Git and exit.',
    '--info-path': 'Print the path where the Info files documenting this version of Git are installed and exit.',
    '-p': 'Pipe all output into less (or if set, $PAGER) if standard output is a terminal.',
    '--paginate': 'Pipe all output into less (or if set, $PAGER) if standard output is a terminal.',
    '--no-pager': 'Do not pipe Git output into a pager.',
    '--git-dir=<path>': 'Set the path to the repository. This can also be controlled by setting the GIT_DIR environment variable.',
    '--work-tree=<path>': 'Set the path to the working tree. It can be an absolute path or a path relative to the current working directory.',
    '--namespace=<path>': 'Set the Git namespace. See gitnamespaces(7) for more details.',
    '--bare': 'Treat the repository as a bare repository. If GIT_DIR environment is not set, it is set to the current working directory.',
    '--no-replace-objects': 'Do not use replacement refs to replace Git objects.',
    '--literal-pathspecs': 'Treat pathspecs literally (i.e. no globbing, no pathspec magic).',
    '--glob-pathspecs': 'Add "glob" magic to all pathspec.',
    '--noglob-pathspecs': 'Add "literal" magic to all pathspec.',
    '--icase-pathspecs': 'Add "icase" magic to all pathspec.',
    'add': 'Add file contents to the index.',
    'am': 'Apply a series of patches from a mailbox.',
    'amend': 'Amend the previous commit.',
    'archive': 'Create an archive of files from a named tree.',
    'bisect': 'Use binary search to find the commit that introduced a bug.',
    'branch': 'List, create, or delete branches.',
    'branches': 'List all branches, including remotes.',
    'bundle': 'Move objects and refs by archive.',
    'c': 'Alias for git clone.',
    'ca': 'Alias for git add -A && git commit -av.',
    'checkout': 'Switch branches or restore working tree files.',
    'cherry-pick': 'Apply the changes introduced by some existing commits.',
    'citool': 'Graphical alternative to git-commit.',
    'clean': 'Remove untracked files from the working tree.',
    'clone': 'Clone a repository into a new directory.',
    'commit': 'Record changes to the repository.',
    'contributors': 'List repo contributors.',
    'describe': 'Describe a commit using the most recent tag reachable from it.',
    'diff': 'Show changes between commits, commit and working tree, etc.',
    'fetch': 'Download objects and refs from another repository.',
    'flow': 'Git branching model extensions.',
    'format-patch': 'Prepare patches for e-mail submission.',
    'gc': 'Cleanup unnecessary files and optimize the local repository.',
    'grep': 'Print lines matching a pattern.',
    'gui': 'A portable graphical interface to Git.',
    'init': 'Create an empty Git repository or reinitialize an existing one.',
    'log': 'Show commit logs.',
    'merge': 'Join two or more development histories together.',
    'mv': 'Move or rename a file, a directory, or a symlink.',
    'notes': 'Add or inspect object notes.',
    'pull': 'Fetch from and integrate with another repository or a local branch.',
    'push': 'Update remote refs along with associated objects.',
    'rebase': 'Forward-port local commits to the updated upstream head.',
    'reset': 'Reset current HEAD to the specified state.',
    'revert': 'Revert some existing commits.',
    'rm': 'Remove files from the working tree and from the index.',
    'shortlog': 'Summarize git log output.',
    'show': 'Show various types of objects.',
    'stage': 'Add file contents to the staging area.',
    'stash': 'Stash the changes in a dirty working directory away.',
    'status': 'Show the working tree status.',
    'submodule': 'Initialize, update or inspect submodules.',
    'tag': 'Create, list, delete or verify a tag object signed with GPG.',
    'tags': 'List all tags.',
    'worktree': 'Manage multiple working trees.',
    'gitk': 'The Git repository browser.',
    'config': 'Get and set repository or global options.',
    'fast-export': 'Git data exporter.',
    'fast-import': 'Backend for fast Git data importers.',
    'filter-branch': 'Rewrite branches.',
    'mergetool': 'Run merge conflict resolution tools to resolve merge conflicts.',
    'pack-refs': 'Pack heads and tags for efficient repository access.',
    'prune': 'Prune all unreachable objects from the object database.',
    'reflog': 'Manage reflog information.',
    'relink': 'Hardlink common objects in local repositories.',
    'remote': 'Manage set of tracked repositories.',
    'remotes': 'List set of tracked repositories.',
    'repack': 'Pack unpacked objects in a repository.',
    'replace': 'Create, list, delete refs to replace objects.',
    'annotate': 'Annotate file lines with commit information.',
    'blame': 'Show what revision and author last modified each line of a file.',
    'cherry': 'Find commits yet to be applied to upstream.',
    'count-objects': 'Count unpacked number of objects and their disk consumption.',
    'difftool': 'Show changes using common diff tools.',
    'fsck': 'Verifies the connectivity and validity of the objects in the database.',
    'get-tar-commit-id': 'Extract commit ID from an archive created using git-archive.',
    'help': 'Display help information about Git.',
    'instaweb': 'Instantly browse your working repository in gitweb.',
    'merge-tree': 'Show three-way merge without touching index.',
    'rerere': 'Reuse recorded resolution of conflicted merges.',
    'rev-parse': 'Pick out and massage parameters.',
    'show-branch': 'Show branches and their commits.',
    'verify-commit': 'Check the GPG signature of commits.',
    'verify-tag': 'Check the GPG signature of tags.',
    'whatchanged': 'Show logs with difference each commit introduces.',
    'gitweb': 'Git web interface (web frontend to Git repositories).',
    'archimport': 'Import an Arch repository into Git.',
    'cvsexportcommit': 'Export a single commit to a CVS checkout.',
    'cvsimport': 'Salvage your data out of another SCM people love to hate.',
    'cvsserver': 'A CVS server emulator for Git.',
    'imap-send': 'Send a collection of patches from stdin to an IMAP folder.',
    'p4': 'Import from and submit to Perforce repositories.',
    'quiltimport': 'Applies a quilt patchset onto the current branch.',
    'request-pull': 'Generates a summary of pending changes.',
    'send-email': 'Send a collection of patches as emails.',
    'svn': 'Bidirectional operation between a Subversion repository and Git.',
    'apply': 'Apply a patch to files and/or to the index.',
    'checkout-index': 'Copy files from the index to the working tree.',
    'commit-tree': 'Create a new commit object.',
    'hash-object': 'Compute object ID and optionally creates a blob from a file.',
    'index-pack': 'Build pack index file for an existing packed archive.',
    'lfs': 'Git extension for versioning large files.',
    'merge-file': 'Run a three-way file merge.',
    'merge-index': 'Run a merge for files needing merging.',
    'mktag': 'Creates a tag object.',
    'mktree': 'Build a tree-object from ls-tree formatted text.',
    'pack-objects': 'Create a packed archive of objects.',
    'prune-packed': 'Remove extra objects that are already in pack files.',
    'read-tree': 'Reads tree information into the index.',
    'symbolic-ref': 'Read, modify and delete symbolic refs.',
    'unpack-objects': 'Unpack objects from a packed archive.',
    'update-index': 'Register file contents in the working tree to the index.',
    'update-ref': 'Update the object name stored in a ref safely.',
    'write-tree': 'Create a tree object from the current index.',
    'cat-file': 'Provide content or type and size information for repository objects.',
    'diff-files': 'Compares files in the working tree and the index.',
    'diff-index': 'Compare a tree to the working tree or index.',
    'diff-tree': 'Compares the content and mode of blobs found via two tree objects.',
    'for-each-ref': 'Output information on each ref.',
    'ls-files': 'Show information about files in the index and the working tree.',
    'ls-remote': 'List references in a remote repository.',
    'ls-tree': 'List the contents of a tree object.',
    'merge-base': 'Find as good common ancestors as possible for a merge.',
    'name-rev': 'Find symbolic names for given revs.',
    'pack-redundant': 'Find redundant pack files.',
    'rev-list': 'Lists commit objects in reverse chronological order.',
    'show-index': 'Show packed archive index.',
    'show-ref': 'List references in a local repository.',
    'unpack-file': "Creates a temporary file with a blob's contents.",
    'var': 'Show a Git logical variable.',
    'verify-pack': 'Validate packed Git archive files.',
    'daemon': 'A really simple server for Git repositories.',
    'fetch-pack': 'Receive missing objects from another repository.',
    'http-backend': 'Server side implementation of Git over HTTP.',
    'send-pack': 'Push objects over Git protocol to another repository.',
    'update-server-info': 'Update auxiliary info file to help dumb servers.',
    'http-fetch': 'Download from a remote Git repository via HTTP.',
    'http-push': 'Push objects over HTTP/DAV to another repository.',
    'parse-remote': 'Routines to help parsing remote repository access parameters.',
    'receive-pack': 'Receive what is pushed into the repository.',
    'shell': 'Restricted login shell for Git-only SSH access.',
    'upload-archive': 'Send archive back to git-archive.',
    'upload-pack': 'Send objects packed back to git-fetch-pack.',
    'check-attr': 'Display gitattributes information.',
    'check-ignore': 'Debug gitignore / exclude files.',
    'check-mailmap': 'Show canonical names and email addresses of contacts.',
    'check-ref-format': 'Ensures that a reference name is well formed.',
    'column': 'Display data in columns.',
    'credential': 'Retrieve and store user credentials.',
    'credential-cache': 'Helper to temporarily store passwords in memory.',
    'credential-store': 'Helper to store credentials on disk.',
    'fmt-merge-msg': 'Produce a merge commit message.',
    'interpret-trailers': 'help add structured information into commit messages.',
    'mailinfo': 'Extracts patch and authorship from a single e-mail message.',
    'mailsplit': 'Simple UNIX mbox splitter program.',
    'merge-one-file': 'The standard helper program to use with git-merge-index.',
    'patch-id': 'Compute unique ID for a patch.',
    'sh-i18n': "Git's i18n setup code for shell scripts.",
    'sh-setup': 'Common Git shell script setup code.',
    'stripspace': 'Remove unnecessary whitespace.',
    'subtree': 'Merge subtrees together and split repository into subtrees.',
    'git-cvsserver': 'A CVS server emulator for Git.',
    'git-credential-osxkeychain': 'Cache GitHub credentials in the OS X Keychain.',
    'git-lfs': 'Git extension for versioning large files.',
    'git-receive-pack': 'Receive what is pushed into the repository',
    'git-shell': 'Restricted login shell for Git-only SSH access.',
    'git-subtree': 'Merge subtrees together and split repository into subtrees.',
    'git-upload-archive': 'Send archive back to git-archive.',
    'git-upload-pack': 'Send objects packed back to git-fetch-pack.',
    'git-flow': 'Git branching model extensions.',
    'gitchangelog': 'Generates a changelog from git tags and commit messages.',
}
META_LOOKUP_GIT_EXTRAS = {
    'git-extras': 'Various Git utilities.',
    'git-alias': 'Define, search and show aliases.',
    'git-archive-file': 'Export the current HEAD of the git repository to a archive.',
    'git-authors': 'Generate authors report.',
    'git-back': 'Undo and Stage latest commits.',
    'git-bug': 'Create bug branch.',
    'git-changelog': 'Generate a changelog report.',
    'git-chore': 'Create chore branch.',
    'git-clear': 'Rigorously clean up a repository.',
    'git-commits-since': 'Show commit logs since some date.',
    'git-contrib': "Show user's contributions",
    'git-count': 'Show commit count.',
    'git-create-branch': 'Create branches.',
    'git-delete-branch': 'Delete branches.',
    'git-delete-merged-branches': 'Delete merged branches.',
    'git-delete-submodule': 'Delete submodules.',
    'git-delete-tag': 'Delete tags.',
    'git-delta': 'Lists changed files.',
    'git-effort': 'Show effort statistics on file(s).',
    'git-feature': 'Create/Merge feature branch.',
    'git-fork': 'Fork a repo on github.',
    'git-fresh-branch': 'Create fresh branches.',
    'git-gh-pages': 'Create the GitHub Pages branch.',
    'git-graft': 'Merge and destroy a given branch.',
    'git-guilt': 'calculate change between two revisions.',
    'git-ignore-io': 'Get sample gitignore file.',
    'git-ignore': 'Add .gitignore patterns.',
    'git-info': 'Returns information on current repository.',
    'git-line-summary': 'Show repository summary by line.',
    'git-local-commits': 'List local commits.',
    'git-lock': 'Lock a file excluded from version control.',
    'git-locked': 'ls files that have been locked.',
    'git-merge-into': 'Merge one branch into another.',
    'git-merge-repo': 'Merge two repo histories.',
    'git-missing': 'Show commits missing from another branch.',
    'git-obliterate': 'Completely remove a file from the repository, including past commits and tags.',
    'git-pr': 'Checks out a pull request locally.',
    'git-psykorebase': 'Rebase a branch with a merge commit.',
    'git-pull-request': 'Checks out a pull request from GitHub.',
    'git-rebase-patch': 'Rebases a patch.',
    'git-refactor': 'Create refactor branch.',
    'git-release': 'Commit, tag and push changes to the repository.',
    'git-rename-tag': 'Rename a tag.',
    'git-repl': 'git read-eval-print-loop.',
    'git-reset-file': 'Reset one file.',
    'git-root': 'show path of root.',
    'git-rscp': 'Copies specific files from the working directory of a remote repository to the current working directory.',
    'git-scp': 'Copy files to SSH compatible git-remote.',
    'git-sed': 'replace patterns in git-controlled files.',
    'git-setup': 'Set up a git repository.',
    'git-show-merged-branches': 'Show merged branches.',
    'git-show-tree': 'show branch tree of commit history.',
    'git-show-unmerged-branches': 'Show unmerged branches.',
    'git-squash': 'Import changes from a branch.',
    'git-summary': 'Show repository summary.',
    'git-touch': 'Touch and add file to the index.',
    'git-undo': 'Remove latest commits.',
    'git-unlock': 'Unlock a file excluded from version control.',

    'extras': 'Various Git utilities.',
    'alias': 'Define, search and show aliases.',
    'archive-file': 'Export the current HEAD of the git repository to a archive.',
    'authors': 'Generate authors report.',
    'back': 'Undo and Stage latest commits.',
    'bug': 'Create bug branch.',
    'changelog': 'Generate a changelog report.',
    'chore': 'Create chore branch.',
    'clear': 'Rigorously clean up a repository.',
    'commits-since': 'Show commit logs since some date.',
    'contrib': "Show user's contributions",
    'count': 'Show commit count.',
    'create-branch': 'Create branches.',
    'delete-branch': 'Delete branches.',
    'delete-merged-branches': 'Delete merged branches.',
    'delete-submodule': 'Delete submodules.',
    'delete-tag': 'Delete tags.',
    'delta': 'Lists changed files.',
    'effort': 'Show effort statistics on file(s).',
    'feature': 'Create/Merge feature branch.',
    'fork': 'Fork a repo on github.',
    'fresh-branch': 'Create fresh branches.',
    'gh-pages': 'Create the GitHub Pages branch.',
    'graft': 'Merge and destroy a given branch.',
    'guilt': 'calculate change between two revisions.',
    'ignore-io': 'Get sample gitignore file.',
    'ignore': 'Add .gitignore patterns.',
    'info': 'Returns information on current repository.',
    'line-summary': 'Show repository summary by line.',
    'local-commits': 'List local commits.',
    'lock': 'Lock a file excluded from version control.',
    'locked': 'ls files that have been locked.',
    'merge-into': 'Merge one branch into another.',
    'merge-repo': 'Merge two repo histories.',
    'missing': 'Show commits missing from another branch.',
    'obliterate': 'Completely remove a file from the repository, including past commits and tags.',
    'pr': 'Checks out a pull request locally.',
    'psykorebase': 'Rebase a branch with a merge commit.',
    'pull-request': 'Checks out a pull request from GitHub.',
    'rebase-patch': 'Rebases a patch.',
    'refactor': 'Create refactor branch.',
    'release': 'Commit, tag and push changes to the repository.',
    'rename-tag': 'Rename a tag.',
    'repl': 'git read-eval-print-loop.',
    'reset-file': 'Reset one file.',
    'root': 'show path of root.',
    'rscp': 'Copies specific files from the working directory of a remote repository to the current working directory.',
    'scp': 'Copy files to SSH compatible git-remote.',
    'sed': 'replace patterns in git-controlled files.',
    'setup': 'Set up a git repository.',
    'show-merged-branches': 'Show merged branches.',
    'show-tree': 'show branch tree of commit history.',
    'show-unmerged-branches': 'Show unmerged branches.',
    'squash': 'Import changes from a branch.',
    'summary': 'Show repository summary.',
    'touch': 'Touch and add file to the index.',
    'undo': 'Remove latest commits.',
    'unlock': 'Unlock a file excluded from version control.',
}



# -*- coding: utf-8 -*-

# Copyright 2015 Donne Martin. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import unicode_literals
from __future__ import print_function

import mock
from tests.compat import unittest

from prompt_toolkit.document import Document

from gitsome.completer import CompleterGitsome


class CompleterTest(unittest.TestCase):

    def setUp(self):
        self.completer = CompleterGitsome()
        self.completer_event = self.create_completer_event()

    def create_completer_event(self):
        return mock.Mock()

    def _get_completions(self, command):
        position = len(command)
        result = self.completer.get_completions(
            Document(text=command, cursor_position=position),
            self.completer_event)
        return result

    def verify_completions(self, commands, expected):
        result = []
        for command in commands:
            # Call the AWS CLI autocompleter
            result.extend(self._get_completions(command))
        result_texts = []
        for item in result:
            # Each result item is a Completion object,
            # we are only interested in the text portion
            result_texts.append(item.text)
        assert result_texts
        if len(expected) == 1:
            assert expected[0] in result_texts
        else:
            for item in expected:
                assert item in result_texts

    def test_blank(self):
        text = ''
        expected = []
        result = self._get_completions(text)
        assert result == expected

    def test_no_completions(self):
        text = 'foo'
        expected = []
        result = self._get_completions(text)
        assert result == expected

    def test_command(self):
        text = ['g']
        expected = ['gh']
        self.verify_completions(text, expected)

    def test_subcommand(self):
        self.verify_completions(['gh c'], ['configure', 'create-comment',
                                           'create-issue', 'create-repo'])
        self.verify_completions(['gh e'], ['emails', 'emojis'])
        self.verify_completions(['gh f'], ['feed', 'followers', 'following'])
        self.verify_completions(['gh g'], ['gitignore-template',
                                           'gitignore-templates'])
        self.verify_completions(['gh i'], ['issue', 'issues'])
        self.verify_completions(['gh l'], ['license', 'licenses'])
        self.verify_completions(['gh m'], ['me'])
        self.verify_completions(['gh n'], ['notifications'])
        self.verify_completions(['gh o'], ['octo'])
        self.verify_completions(['gh p'], ['pull-request', 'pull-requests'])
        self.verify_completions(['gh r'], ['rate-limit', 'repo', 'repos'])
        self.verify_completions(['gh s'], ['search-issues', 'search-repos',
                                           'starred'])
        self.verify_completions(['gh t'], ['trending'])
        self.verify_completions(['gh u'], ['user'])
        self.verify_completions(['gh v'], ['view'])

    def test_args(self):
        self.verify_completions(['gh octo '], ['"Keep it logically awesome"'])

    def test_no_args_with_options(self):
        self.verify_completions(['gh octo '], ['"Keep it logically awesome"'])

    def test_options(self):
        self.verify_completions(['gh emojis '],
                                ['-p', '--pager'])

    def test_multiple_options(self):
        self.verify_completions(['gh feed -pr --pa'], ['--pager'])
        self.verify_completions(['gh feed foo -pr --pa'], ['--pager'])

    def test_fuzzy(self):
        self.completer.fuzzy_match = True
        self.verify_completions(['gh ot'], ['octo'])

    def test_build_completions_with_meta(self):
        result = self.completer.build_completions_with_meta('git ad',
                                                            'ad',
                                                            ['add'])
        assert result[0].display_meta_text == 'Add file contents to the index.'
        result = self.completer.build_completions_with_meta('git-alia',
                                                            'git-alia',
                                                            ['git-alias'])
        assert result[0].display_meta_text == 'Define, search and show aliases.'



# -*- coding: utf-8 -*-
"""
github3.gists.comment
---------------------

Module containing the logic for a GistComment

"""
from __future__ import unicode_literals

from ..models import BaseComment
from ..users import User


class GistComment(BaseComment):

    """This object represents a comment on a gist.

    Two comment instances can be checked like so::

        c1 == c2
        c1 != c2

    And is equivalent to::

        c1.id == c2.id
        c1.id != c2.id

    See also: http://developer.github.com/v3/gists/comments/

    """

    def _update_attributes(self, comment):
        self._api = comment.get('url')
        #: :class:`User <github3.users.User>` who made the comment
        #: Unless it is not associated with an account
        self.user = None
        if comment.get('user'):
            self.user = User(comment.get('user'), self)  # (No coverage)

    def _repr(self):
        return '<Gist Comment [{0}]>'.format(self.user.login)



# -*- coding: utf-8 -*-
"""
github3.orgs
============

This module contains all of the classes related to organizations.

"""
from __future__ import unicode_literals

import warnings

from json import dumps
from .events import Event
from .models import BaseAccount, GitHubCore
from .repos import Repository
from .users import User
from .decorators import requires_auth
from uritemplate import URITemplate


class Team(GitHubCore):

    """The :class:`Team <Team>` object.

    Two team instances can be checked like so::

        t1 == t2
        t1 != t2

    And is equivalent to::

        t1.id == t2.id
        t1.id != t2.id

    See also: http://developer.github.com/v3/orgs/teams/

    """

    # Roles available to members on a team.
    members_roles = frozenset(['member', 'maintainer', 'all'])

    def _update_attributes(self, team):
        self._api = team.get('url', '')
        #: This team's name.
        self.name = team.get('name')
        #: Unique ID of the team.
        self.id = team.get('id')
        #: Permission level of the group.
        self.permission = team.get('permission')
        #: Number of members in this team.
        self.members_count = team.get('members_count')
        members = team.get('members_url')
        #: Members URL Template. Expands with ``member``.
        self.members_urlt = URITemplate(members) if members else None
        #: Number of repos owned by this team.
        self.repos_count = team.get('repos_count')
        #: Repositories url (not a template).
        self.repositories_url = team.get('repositories_url')

    def _repr(self):
        return '<Team [{0}]>'.format(self.name)

    @requires_auth
    def add_member(self, username):
        """Add ``username`` to this team.

        :param str username: the username of the user you would like to add to
            the team.
        :returns: bool
        """
        warnings.warn(
            'This is no longer supported by the GitHub API, see '
            'https://developer.github.com/changes/2014-09-23-one-more-week'
            '-before-the-add-team-member-api-breaking-change/',
            DeprecationWarning)
        url = self._build_url('members', username, base_url=self._api)
        return self._boolean(self._put(url), 204, 404)

    @requires_auth
    def add_repository(self, repository):
        """Add ``repository`` to this team.

        :param str repository: (required), form: 'user/repo'
        :returns: bool
        """
        url = self._build_url('repos', repository, base_url=self._api)
        return self._boolean(self._put(url), 204, 404)

    @requires_auth
    def delete(self):
        """Delete this team.

        :returns: bool
        """
        return self._boolean(self._delete(self._api), 204, 404)

    @requires_auth
    def edit(self, name, permission=''):
        """Edit this team.

        :param str name: (required)
        :param str permission: (optional), ('pull', 'push', 'admin')
        :returns: bool
        """
        if name:
            data = {'name': name, 'permission': permission}
            json = self._json(self._patch(self._api, data=dumps(data)), 200)
            if json:
                self._update_attributes(json)
                return True
        return False

    @requires_auth
    def has_repository(self, repository):
        """Check if this team has access to ``repository``.

        :param str repository: (required), form: 'user/repo'
        :returns: bool
        """
        url = self._build_url('repos', repository, base_url=self._api)
        return self._boolean(self._get(url), 204, 404)

    @requires_auth
    def invite(self, username):
        """Invite the user to join this team.

        This returns a dictionary like so::

            {'state': 'pending', 'url': 'https://api.github.com/teams/...'}

        :param str username: (required), user to invite to join this team.
        :returns: dictionary
        """
        url = self._build_url('memberships', username, base_url=self._api)
        return self._json(self._put(url), 200)

    @requires_auth
    def is_member(self, username):
        """Check if ``login`` is a member of this team.

        :param str username: (required), username name of the user
        :returns: bool
        """
        url = self._build_url('members', username, base_url=self._api)
        return self._boolean(self._get(url), 204, 404)

    @requires_auth
    def members(self, role=None, number=-1, etag=None):
        r"""Iterate over the members of this team.

        :param str role: (optional), filter members returned by their role
            in the team. Can be one of: ``"member"``, ``"maintainer"``,
            ``"all"``. Default: ``"all"``.
        :param int number: (optional), number of users to iterate over.
            Default: -1 iterates over all values
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of :class:`User <github3.users.User>`\ s
        """
        headers = {}
        params = {}
        if role in self.members_roles:
            params['role'] = role
            headers['Accept'] = 'application/vnd.github.ironman-preview+json'
        url = self._build_url('members', base_url=self._api)
        return self._iter(int(number), url, User, params=params, etag=etag,
                          headers=headers)

    @requires_auth
    def repositories(self, number=-1, etag=None):
        """Iterate over the repositories this team has access to.

        :param int number: (optional), number of repos to iterate over.
            Default: -1 iterates over all values
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of :class:`Repository <github3.repos.Repository>`
            objects
        """
        headers = {'Accept': 'application/vnd.github.ironman-preview+json'}
        url = self._build_url('repos', base_url=self._api)
        return self._iter(int(number), url, Repository, etag=etag,
                          headers=headers)

    @requires_auth
    def membership_for(self, username):
        """Retrieve the membership information for the user.

        :param str username: (required), name of the user
        :returns: dictionary
        """
        url = self._build_url('memberships', username, base_url=self._api)
        json = self._json(self._get(url), 200)
        return json or {}

    @requires_auth
    def remove_member(self, username):
        """Remove ``username`` from this team.

        :param str username: (required), username of the member to remove
        :returns: bool
        """
        warnings.warn(
            'This is no longer supported by the GitHub API, see '
            'https://developer.github.com/changes/2014-09-23-one-more-week'
            '-before-the-add-team-member-api-breaking-change/',
            DeprecationWarning)
        url = self._build_url('members', username, base_url=self._api)
        return self._boolean(self._delete(url), 204, 404)

    @requires_auth
    def revoke_membership(self, username):
        """Revoke this user's team membership.

        :param str username: (required), name of the team member
        :returns: bool
        """
        url = self._build_url('memberships', username, base_url=self._api)
        return self._boolean(self._delete(url), 204, 404)

    @requires_auth
    def remove_repository(self, repository):
        """Remove ``repository`` from this team.

        :param str repository: (required), form: 'user/repo'
        :returns: bool
        """
        url = self._build_url('repos', repository, base_url=self._api)
        return self._boolean(self._delete(url), 204, 404)


class Organization(BaseAccount):

    """The :class:`Organization <Organization>` object.

    Two organization instances can be checked like so::

        o1 == o2
        o1 != o2

    And is equivalent to::

        o1.id == o2.id
        o1.id != o2.id

    See also: http://developer.github.com/v3/orgs/

    """

    # Filters available when listing members. Note: ``"2fa_disabled"``
    # is only available for organization owners.
    members_filters = frozenset(['2fa_disabled', 'all'])

    # Roles available to members in an organization.
    members_roles = frozenset(['all', 'admin', 'member'])

    def _update_attributes(self, org):
        super(Organization, self)._update_attributes(org)
        self.type = self.type or 'Organization'

        #: Events url (not a template)
        self.events_url = org.get('events_url')
        #: Number of private repositories.
        self.private_repos = org.get('private_repos', 0)

        members = org.get('members_url')
        #: Members URL Template. Expands with ``member``
        self.members_urlt = URITemplate(members) if members else None

        members = org.get('public_members_url')
        #: Public Members URL Template. Expands with ``member``
        self.public_members_urlt = URITemplate(members) if members else None
        #: Repositories url (not a template)
        self.repos_url = org.get('repos_url')

    @requires_auth
    def add_member(self, username, team_id):
        """Add ``username`` to ``team`` and thereby to this organization.

        .. warning::
            This method is no longer valid. To add a member to a team, you
            must now retrieve the team directly, and use the ``invite``
            method.

        .. warning::
            This method is no longer valid. To add a member to a team, you
            must now retrieve the team directly, and use the ``invite``
            method.

        Any user that is to be added to an organization, must be added
        to a team as per the GitHub api.

        .. versionchanged:: 1.0

            The second parameter used to be ``team`` but has been changed to
            ``team_id``. This parameter is now required to be an integer to
            improve performance of this method.

        :param str username: (required), login name of the user to be added
        :param int team_id: (required), team id
        :returns: bool
        """
        warnings.warn(
            'This is no longer supported by the GitHub API, see '
            'https://developer.github.com/changes/2014-09-23-one-more-week'
            '-before-the-add-team-member-api-breaking-change/',
            DeprecationWarning)

        if int(team_id) < 0:
            return False

        url = self._build_url('teams', str(team_id), 'members', str(username))
        return self._boolean(self._put(url), 204, 404)

    @requires_auth
    def add_repository(self, repository, team_id):
        """Add ``repository`` to ``team``.

        .. versionchanged:: 1.0

            The second parameter used to be ``team`` but has been changed to
            ``team_id``. This parameter is now required to be an integer to
            improve performance of this method.

        :param str repository: (required), form: 'user/repo'
        :param int team_id: (required), team id
        :returns: bool
        """
        if int(team_id) < 0:
            return False

        url = self._build_url('teams', str(team_id), 'repos', str(repository))
        return self._boolean(self._put(url), 204, 404)

    @requires_auth
    def create_repository(self, name, description='', homepage='',
                          private=False, has_issues=True, has_wiki=True,
                          team_id=0, auto_init=False, gitignore_template='',
                          license_template=''):
        """Create a repository for this organization.

        If the client is authenticated and a member of the organization, this
        will create a new repository in the organization.

        :param str name: (required), name of the repository
        :param str description: (optional)
        :param str homepage: (optional)
        :param bool private: (optional), If ``True``, create a private
            repository. API default: ``False``
        :param bool has_issues: (optional), If ``True``, enable issues for
            this repository. API default: ``True``
        :param bool has_wiki: (optional), If ``True``, enable the wiki for
            this repository. API default: ``True``
        :param int team_id: (optional), id of the team that will be granted
            access to this repository
        :param bool auto_init: (optional), auto initialize the repository.
        :param str gitignore_template: (optional), name of the template; this
            is ignored if auto_int = False.
        :param str license_template: (optional), name of the license; this
            is ignored if auto_int = False.
        :returns: :class:`Repository <github3.repos.Repository>`

        .. warning: ``name`` should be no longer than 100 characters
        """
        url = self._build_url('repos', base_url=self._api)
        data = {'name': name, 'description': description,
                'homepage': homepage, 'private': private,
                'has_issues': has_issues, 'has_wiki': has_wiki,
                'license_template': license_template, 'auto_init': auto_init,
                'gitignore_template': gitignore_template}
        if int(team_id) > 0:
            data.update({'team_id': team_id})
        json = self._json(self._post(url, data), 201)
        return self._instance_or_null(Repository, json)

    @requires_auth
    def conceal_member(self, username):
        """Conceal ``username``'s membership in this organization.

        :param str username: username of the organization member to conceal
        :returns: bool
        """
        url = self._build_url('public_members', username, base_url=self._api)
        return self._boolean(self._delete(url), 204, 404)

    @requires_auth
    def create_team(self, name, repo_names=[], permission=''):
        """Create a new team and return it.

        This only works if the authenticated user owns this organization.

        :param str name: (required), name to be given to the team
        :param list repo_names: (optional) repositories, e.g.
            ['github/dotfiles']
        :param str permission: (optional), options:

            - ``pull`` -- (default) members can not push or administer
                repositories accessible by this team
            - ``push`` -- members can push and pull but not administer
                repositories accessible by this team
            - ``admin`` -- members can push, pull and administer
                repositories accessible by this team

        :returns: :class:`Team <Team>`
        """
        data = {'name': name, 'repo_names': repo_names,
                'permission': permission}
        url = self._build_url('teams', base_url=self._api)
        json = self._json(self._post(url, data), 201)
        return self._instance_or_null(Team, json)

    @requires_auth
    def edit(self, billing_email=None, company=None, email=None, location=None,
             name=None):
        """Edit this organization.

        :param str billing_email: (optional) Billing email address (private)
        :param str company: (optional)
        :param str email: (optional) Public email address
        :param str location: (optional)
        :param str name: (optional)
        :returns: bool
        """
        json = None
        data = {'billing_email': billing_email, 'company': company,
                'email': email, 'location': location, 'name': name}
        self._remove_none(data)

        if data:
            json = self._json(self._patch(self._api, data=dumps(data)), 200)

        if json:
            self._update_attributes(json)
            return True
        return False

    def is_member(self, username):
        """Check if the user named ``username`` is a member.

        :param str username: name of the user you'd like to check
        :returns: bool
        """
        url = self._build_url('members', username, base_url=self._api)
        return self._boolean(self._get(url), 204, 404)

    def is_public_member(self, username):
        """Check if the user named ``username`` is a public member.

        :param str username: name of the user you'd like to check
        :returns: bool
        """
        url = self._build_url('public_members', username, base_url=self._api)
        return self._boolean(self._get(url), 204, 404)

    def events(self, number=-1, etag=None):
        r"""Iterate over events for this org.

        :param int number: (optional), number of events to return. Default: -1
            iterates over all events available.
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of :class:`Event <github3.events.Event>`\ s
        """
        url = self._build_url('events', base_url=self._api)
        return self._iter(int(number), url, Event, etag=etag)

    def members(self, filter=None, role=None, number=-1, etag=None):
        r"""Iterate over members of this organization.

        :param str filter: (optional), filter members returned by this method.
            Can be one of: ``"2fa_disabled"``, ``"all",``. Default: ``"all"``.
            Filtering by ``"2fa_disabled"`` is only available for organization
            owners with private repositories.
        :param str role: (optional), filter members returned by their role.
            Can be one of: ``"all"``, ``"admin"``, ``"member"``. Default:
            ``"all"``.
        :param int number: (optional), number of members to return. Default:
            -1 will return all available.
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of :class:`User <github3.users.User>`\ s
        """
        headers = {}
        params = {}
        if filter in self.members_filters:
            params['filter'] = filter
        if role in self.members_roles:
            params['role'] = role
            headers['Accept'] = 'application/vnd.github.ironman-preview+json'
        url = self._build_url('members', base_url=self._api)
        return self._iter(int(number), url, User, params=params, etag=etag,
                          headers=headers)

    def public_members(self, number=-1, etag=None):
        r"""Iterate over public members of this organization.

        :param int number: (optional), number of members to return. Default:
            -1 will return all available.
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of :class:`User <github3.users.User>`\ s
        """
        url = self._build_url('public_members', base_url=self._api)
        return self._iter(int(number), url, User, etag=etag)

    def repositories(self, type='', number=-1, etag=None):
        r"""Iterate over repos for this organization.

        :param str type: (optional), accepted values:
            ('all', 'public', 'member', 'private', 'forks', 'sources'), API
            default: 'all'
        :param int number: (optional), number of members to return. Default:
            -1 will return all available.
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of :class:`Repository <github3.repos.Repository>`
        """
        url = self._build_url('repos', base_url=self._api)
        params = {}
        if type in ('all', 'public', 'member', 'private', 'forks', 'sources'):
            params['type'] = type
        return self._iter(int(number), url, Repository, params, etag)

    @requires_auth
    def teams(self, number=-1, etag=None):
        r"""Iterate over teams that are part of this organization.

        :param int number: (optional), number of teams to return. Default: -1
            returns all available teams.
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of :class:`Team <Team>`\ s
        """
        url = self._build_url('teams', base_url=self._api)
        return self._iter(int(number), url, Team, etag=etag)

    @requires_auth
    def publicize_member(self, username):
        """Make ``username``'s membership in this organization public.

        :param str username: the name of the user whose membership you wish to
            publicize
        :returns: bool
        """
        url = self._build_url('public_members', username, base_url=self._api)
        return self._boolean(self._put(url), 204, 404)

    @requires_auth
    def remove_member(self, username):
        """Remove the user named ``username`` from this organization.

        :param str username: name of the user to remove from the org
        :returns: bool
        """
        url = self._build_url('members', username, base_url=self._api)
        return self._boolean(self._delete(url), 204, 404)

    @requires_auth
    def remove_repository(self, repository, team_id):
        """Remove ``repository`` from the team with ``team_id``.

        :param str repository: (required), form: 'user/repo'
        :param int team_id: (required)
        :returns: bool
        """
        if int(team_id) > 0:
            url = self._build_url('teams', str(team_id), 'repos',
                                  str(repository))
            return self._boolean(self._delete(url), 204, 404)
        return False

    @requires_auth
    def team(self, team_id):
        """Return the team specified by ``team_id``.

        :param int team_id: (required), unique id for the team
        :returns: :class:`Team <Team>`
        """
        json = None
        if int(team_id) > 0:
            url = self._build_url('teams', str(team_id))
            json = self._json(self._get(url), 200)
        return self._instance_or_null(Team, json)


class Membership(GitHubCore):

    """The wrapper for information about Team and Organization memberships."""

    def _repr(self):
        return '<Membership [{0}]>'.format(self.organization)

    def _update_attributes(self, membership):
        self._api = membership.get('url')
        self.organization = Organization(membership.get('organization', {}),
                                         self)
        self.state = membership.get('state', '')
        self.organization_url = membership.get('organization_url')
        self.active = self.state.lower() == 'active'
        self.pending = self.state.lower() == 'pending'

    @requires_auth
    def edit(self, state):
        """Edit the user's membership.

        :param str state: (required), the state the membership should be in.
            Only accepts ``"active"``.
        :returns: whether the edit was successful or not
        :rtype: bool
        """
        if state and state.lower() == 'active':
            data = dumps({'state': state.lower()})
            json = self._json(self._patch(self._api, data=data))
            self._update_attributes(json)
            return True
        return False



# amalgamate exclude
import os as _os

if _os.getenv("XONSH_DEBUG", ""):
    pass
else:
    import sys as _sys

    try:
        from xonsh.prompt import __amalgam__

        cwd = __amalgam__
        _sys.modules["xonsh.prompt.cwd"] = __amalgam__
        env = __amalgam__
        _sys.modules["xonsh.prompt.env"] = __amalgam__
        gitstatus = __amalgam__
        _sys.modules["xonsh.prompt.gitstatus"] = __amalgam__
        job = __amalgam__
        _sys.modules["xonsh.prompt.job"] = __amalgam__
        vc = __amalgam__
        _sys.modules["xonsh.prompt.vc"] = __amalgam__
        base = __amalgam__
        _sys.modules["xonsh.prompt.base"] = __amalgam__
        del __amalgam__
    except ImportError:
        pass
    del _sys
del _os
# amalgamate end



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from ..models import GitHubCore
from ..repos import Repository


class CodeSearchResult(GitHubCore):
    def _update_attributes(self, data):
        self._api = data.get('url')
        #: Filename the match occurs in
        self.name = data.get('name')
        #: Path in the repository to the file
        self.path = data.get('path')
        #: SHA in which the code can be found
        self.sha = data.get('sha')
        #: URL to the Git blob endpoint
        self.git_url = data.get('git_url')
        #: URL to the HTML view of the blob
        self.html_url = data.get('html_url')
        #: Repository the code snippet belongs to
        self.repository = Repository(data.get('repository', {}), self)
        #: Score of the result
        self.score = data.get('score')
        #: Text matches
        self.text_matches = data.get('text_matches', [])

    def _repr(self):
        return '<CodeSearchResult [{0}]>'.format(self.path)
